#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the csiactdata script
#
# Copyright (C) 2016-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
#import gammalib
import cscripts
from testing import test


# ================================ #
# Test class for csiactdata script #
# ================================ #
class Test(test):
    """
    Test class for csiactdata script

    This test class makes unit tests for the csiactdata script by using it
    from the command line and from Python.
    """
    
    # Constructor
    def __init__(self):
        """
        Constructor.
        """
        # Call base class constructor
        test.__init__(self)

        # Set data members
        self._datapath = self._datadir + '/iactdata'

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('csiactdata')

        # Append tests
        self.append(self._test_cmd, 'Test csiactdata on command line')
        self.append(self._test_python, 'Test csiactdata from Python')

        # Return
        return

    # Test csiactdata on command line
    def _test_cmd(self):
        """
        Test csiactdata on the command line
        """
        # Set script name
        csiactdata = self._script('csiactdata')

        # Setup csiactdata command
        cmd = csiactdata+' datapath="'+self._datapath+'"'+ \
                         ' logfile="csiactdata_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Setup csiactdata command
        cmd = csiactdata+' datapath="data_path_that_does_not_exist"'+ \
                         ' logfile="csiactdata_cmd2.log" debug=yes'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check csiactdata --help
        self._check_help(csiactdata)

        # Return
        return

    # Test csiactdata from Python
    def _test_python(self):
        """
        Test csiactdata from Python
        """
        # Set-up csiactdata
        iactdata = cscripts.csiactdata()
        iactdata['datapath'] = self._datapath
        iactdata['logfile']  = 'csiactdata_py1.log'
        iactdata['chatter']  = 2

        # Run csiactdata script and save run list
        iactdata.logFileOpen()   # Make sure we get a log file
        iactdata.run()
        
        # Get available configs
        names = iactdata.names()
        
        # Check if there is at least one config available
        self.test_assert(len(names) > 0,
             'Check successful execution from Python')

        # Set-up csiactdata
        iactdata = cscripts.csiactdata()
        iactdata['datapath'] = self._datapath
        iactdata['logfile']  = 'csiactdata_py2.log'
        iactdata['chatter']  = 4

        # Run csiactdata script and save run list
        iactdata.execute()
        
        # Get available configs
        names = iactdata.names()
        
        # Check if there is at least one config available
        self.test_assert(len(names) > 0,
             'Check successful execution from Python')

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the csobsdef script.
#
# Copyright (C) 2016-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import gammalib
import cscripts
from testing import test


# ============================== #
# Test class for csobsdef script #
# ============================== #
class Test(test):
    """
    Test class for csobsdef script

    This test class makes unit tests for the csobsdef script by using it
    from the command line and from Python.
    """
    
    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Set members
        self._pntdef_min = self._datadir + '/pntdef_minimal.dat'
        self._pntdef_max = self._datadir + '/pntdef_maximal.dat'

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('csobsdef')

        # Append tests
        self.append(self._test_cmd, 'Test csobsdef on command line')
        self.append(self._test_python, 'Test csobsdef from Python')

        # Return
        return

    # Test csobsdef on command line
    def _test_cmd(self):
        """
        Test csobsdef on the command line
        """
        # Set script name
        csobsdef = self._script('csobsdef')

        # Setup csobsdef command
        cmd = csobsdef+' inpnt="'+self._pntdef_min+'"'+ \
                       ' outobs="csobsdef_cmd1.xml"'+ \
                       ' caldb="'+self._caldb+'" irf="'+self._irf+'"'+ \
                       ' emin=0.1 emax=100.0 duration=1800.0 rad=5.0'+ \
                       ' logfile="csobsdef_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check observation definition file
        self._check_obsdef_file('csobsdef_cmd1.xml')

        # Setup csobsdef command
        cmd = csobsdef+' inpnt="obs_definition_that_does_not_exist.dat"'+ \
                       ' outobs="csobsdef_cmd1.xml"'+ \
                       ' caldb="'+self._caldb+'" irf="'+self._irf+'"'+ \
                       ' emin=0.1 emax=100.0 duration=1800.0 rad=5.0'+ \
                       ' logfile="csobsdef_cmd2.log" debug=yes'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check csobsdef --help
        self._check_help(csobsdef)

        # Return
        return

    # Test csobsdef from Python
    def _test_python(self):
        """
        Test csobsdef from Python
        """
        # Set-up csobsdef with minimal pointing definition file
        obsdef = cscripts.csobsdef()
        obsdef['inpnt']    = self._pntdef_min
        obsdef['outobs']   = 'csobsdef_py1.xml'
        obsdef['caldb']    = self._caldb
        obsdef['irf']      = self._irf
        obsdef['emin']     = 0.1
        obsdef['emax']     = 100.0
        obsdef['duration'] = 1800.0
        obsdef['rad']      = 5.0
        obsdef['logfile']  = 'csobsdef_py1.log'
        obsdef['chatter']  = 2

        # Run csobsdef script and save XML file
        obsdef.logFileOpen()   # Make sure we get a log file
        obsdef.run()
        obsdef.save()

        # Check model file
        self._check_obsdef_file('csobsdef_py1.xml')

        # Set-up csobsdef with maximal pointing definition file
        obsdef = cscripts.csobsdef()
        obsdef['inpnt']    = self._pntdef_max
        obsdef['outobs']   = 'csobsdef_py2.xml'
        obsdef['logfile']  = 'csobsdef_py2.log'
        obsdef['chatter']  = 3

        # Execute csobsdef script
        obsdef.execute()

        # Check model file
        self._check_obsdef_file('csobsdef_py2.xml')

        # Load CSV table
        csv = gammalib.GCsv(self._pntdef_min, ',')
 
        # Set-up csobsdef
        obsdef = cscripts.csobsdef()
        obsdef.pntdef(csv)
        obsdef['outobs']   = 'csobsdef_py3.xml'
        obsdef['caldb']    = self._caldb
        obsdef['irf']      = self._irf
        obsdef['emin']     = 0.1
        obsdef['emax']     = 100.0
        obsdef['duration'] = 1800.0
        obsdef['rad']      = 5.0
        obsdef['logfile']  = 'csobsdef_py3.log'
        obsdef['chatter']  = 4

        # Execute csobsdef script
        obsdef.execute()

        # Check model file
        self._check_obsdef_file('csobsdef_py3.xml')

        # Return
        return

    # Check observation definition file
    def _check_obsdef_file(self, filename):
        """
        Check observation definition file
        """
        # Load observation definition file
        obs = gammalib.GObservations(filename)

        # Check number of observations
        self.test_value(obs.size(), 3,
             'Check for 3 observations in observation definition file')

        # Loop over all observations
        for o in obs:
            self.test_value(o.instrument(), 'CTA',
                            'Check for "CTA" instrument')
            self.test_value(o.ontime(), 1800.0, 1.0e-6,
                            'Check for ontime')
            self.test_value(o.deadc(), 0.98, 1.0e-6,
                            'Check for deadtime correction')
            self.test_value(o.events().ebounds().emin().TeV(), 0.1, 1.0e-6,
                            'Check for minimum energy')
            self.test_value(o.events().ebounds().emax().TeV(), 100.0, 1.0e-6,
                            'Check for maximum energy')
            self.test_value(o.roi().radius(), 5.0, 1.0e-6,
                            'Check for ROI radius')

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# Show the content of the ctools calibration database
#
# Copyright (C) 2014-2017 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import glob
import os
import gammalib
import ctools


# ============= #
# cscaldb class #
# ============= #
class cscaldb(ctools.cscript):
    """
    Shows the content of the ctools calibration database.

    The ``cscaldb`` script writes the content of the calibration database
    into the ``cscaldb.log`` log file. If the ``debug`` parameter is set
    to ``yes`` the calibration database is also logged in the console.
    """

    # Constructor
    def __init__(self, *argv):
        """
        Constructor.

        Parameters
        ----------
        argv : list of str
            List of IRAF command line parameter strings of the form
            ``parameter=3``.

        Raises
        ------
        TypeError
            An invalid number of command line arguments was provided.
        """
        # Initialise application by calling the base class constructor
        self._init_cscript(self.__class__.__name__, ctools.__version__, argv)

        # Return
        return


    # Private methods
    #
    # Query all user parameters
    def _get_parameters(self):

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)

        # Return
        return

    # Extract mission names from a calibration database
    def _get_missions(self, caldb):

        # Initialise mission list
        missions = []

        # Extract missions
        paths = glob.glob(caldb.rootdir()+'/data/*')
        paths.sort()
        for path in paths:
            missions.append(os.path.basename(path))

        # Sort missions
        missions.sort()

        # Return missions
        return missions

    # Extract instrument names from a calibration database
    def _get_instruments(self, caldb, mission):

        # Initialise instrument list
        instruments = []

        # Extract instruments
        paths       = glob.glob(caldb.rootdir()+'/data/'+mission+'/*')
        paths.sort()
        for path in paths:
            instruments.append(os.path.basename(path))

        # Sort instruments
        instruments.sort()

        # Return instruments
        return instruments

    # Extract response names from a calibrations FITS table and return them
    # in form of a list
    def _get_response_names(self, calibrations):

        # Initialise response name list
        names = []

        # Extract response names from calibrations and append them to the
        # response name list
        nrows = calibrations.nrows()
        ncols = calibrations.number()
        for row in range(nrows):
            for col in range(ncols):
                cal    = calibrations.string(row, col)
                istart = cal.find('NAME(')
                if istart != -1:
                    istop = cal.find(')')
                    name  = cal[5:istop]
                    if names.count(name) == 0:
                        names.append(name)

        # Sort response names
        names.sort()

        # Return response names
        return names


    # Public methods
    def run(self):
        """
        Run the script.
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()

        # Get the calibration database
        caldb = gammalib.GCaldb()

        # Extract mission names from the calibration database
        missions = self._get_missions(caldb)

        # Loop over missions
        for mission in missions:

            # Skip all non-CTA instruments
            if mission != 'cta':
                continue

            # Write mission into logger
            self._log_header1(gammalib.TERSE, 'Mission: '+mission)

            # Extract instruments
            instruments = self._get_instruments(caldb, mission)

            # Loop over instruments
            for instrument in instruments:

                # Write instrument into logger
                self._log_header3(gammalib.TERSE, 'Response functions in database "'+
                                                  instrument+'"')

                # Open calibration index file and retrieve calibrations
                filename = '/data/'+mission+'/'+instrument+'/caldb.indx'
                cifname  = caldb.rootdir() + filename
                fits     = gammalib.GFits(cifname)
                cif      = fits['CIF']
                caltable = cif['CAL_CBD']

                # Extract response names
                names = self._get_response_names(caltable)

                # Print response name
                if self._logTerse():
                    for name in names:
                        self._log(name+'\n')
                    self._log('\n')

        # Return
        return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = cscaldb(sys.argv)

    # Execute application
    app.execute()



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the csiactcopy script.
#
# Copyright (C) 2016-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import os
import gammalib
import cscripts
from testing import test


# ================================ #
# Test class for csiactcopy script #
# ================================ #
class Test(test):
    """
    Test class for csiactcopy script

    This test class makes unit tests for the csiactcopy script by using it
    from the command line and from Python.
    """
    
    # Constructor
    def __init__(self):
        """
        Constructor.
        """
        # Call base class constructor
        test.__init__(self)

        # Set data members
        self._remote   = self._datadir + '/iactdata/master.json'
        self._runlist1 = self._datadir + '/iact_runlist_1.dat'
        self._runlist2 = self._datadir + '/iact_runlist_2.dat'

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('csiactcopy')

        # Append tests
        self.append(self._test_cmd, 'Test csiactcopy on command line')
        self.append(self._test_python, 'Test csiactcopy from Python')

        # Return
        return

    # Test csiactcopy on command line
    def _test_cmd(self):
        """
        Test csiactcopy on the command line
        """
        # Set script name
        csiactcopy = self._script('csiactcopy')

        # Setup csiactcopy command
        cmd = csiactcopy+' remote_master="'+self._remote+'"'+ \
                         ' prodname="unit-test"'+ \
                         ' outpath="csiactcopy_cmd1"'+ \
                         ' logfile="csiactcopy_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check copy
        self._check_copy('csiactcopy_cmd1')

        # Setup csiactcopy command
        cmd = csiactcopy+' remote_master="master_that_does_not_exist"'+ \
                         ' prodname="unit-test"'+ \
                         ' outpath="csiactcopy_cmd2"'+ \
                         ' logfile="csiactcopy_cmd2.log" debug=yes'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check csiactcopy --help
        self._check_help(csiactcopy)

        # Return
        return

    # Test csiactcopy from Python
    def _test_python(self):
        """
        Test csiactcopy from Python
        """
        # Set-up csiactcopy
        iactcopy = cscripts.csiactcopy()
        iactcopy['remote_master'] = self._remote
        iactcopy['prodname']      = 'unit-test'
        iactcopy['outpath']       = 'csiactcopy_py1'
        iactcopy['logfile']       = 'csiactcopy_py1.log'
        iactcopy['chatter']       = 2

        # Run csiactcopy script and save run list
        iactcopy.logFileOpen()   # Make sure we get a log file
        iactcopy.run()
        iactcopy.save()
        
        # Check IACT data store
        self._check_copy('csiactcopy_py1')
        self._check_n_obs('csiactcopy_py1', 9)
        
        # Set-up csiactcopy with clobber=True, should lead to skipping
        # of files
        iactcopy = cscripts.csiactcopy()
        iactcopy['remote_master'] = self._remote
        iactcopy['prodname']      = 'unit-test'
        iactcopy['outpath']       = 'csiactcopy_py1'
        iactcopy['logfile']       = 'csiactcopy_py2.log'
        iactcopy['chatter']       = 3
        iactcopy['clobber']       = True

        # Execute csiactcopy script
        iactcopy.logFileOpen()   # Make sure we get a log file
        iactcopy.execute()
        
        # Check copy
        self._check_copy('csiactcopy_py1')
        self._check_n_obs('csiactcopy_py1', 9)
        
        # Set-up csiactcopy with clobber=False, should also lead to skipping
        # of files, but for the reason that it's not allowed to overwrite
        # the file
        iactcopy = cscripts.csiactcopy()
        iactcopy['remote_master'] = self._remote
        iactcopy['prodname']      = 'unit-test'
        iactcopy['outpath']       = 'csiactcopy_py1'
        iactcopy['logfile']       = 'csiactcopy_py3.log'
        iactcopy['chatter']       = 4
        iactcopy['clobber']       = False

        # Run csiactcopy script and save run list
        iactcopy.logFileOpen()   # Make sure we get a log file
        iactcopy.run()
        iactcopy.save()
        
        # Check copy
        self._check_copy('csiactcopy_py1')
        self._check_n_obs('csiactcopy_py1', 9)
        
        # Remove the copied IACT data store before continuing
        if os.path.isdir('csiactcopy_py2'):
            os.system('rm -r csiactcopy_py2')
        
        # Set-up csiactcopy using a run list with 5 runs
        iactcopy = cscripts.csiactcopy()
        iactcopy['runlist']       = self._runlist1
        iactcopy['remote_master'] = self._remote
        iactcopy['prodname']      = 'unit-test'
        iactcopy['outpath']       = 'csiactcopy_py2'
        iactcopy['logfile']       = 'csiactcopy_py4.log'
        iactcopy['chatter']       = 4

        # Run csiactcopy script and save run list
        iactcopy.logFileOpen()   # Make sure we get a log file
        iactcopy.run()
        iactcopy.save()

        # Check copy
        self._check_copy('csiactcopy_py2')
        self._check_n_obs('csiactcopy_py2', 5)

        # Set-up csiactcopy using another run list with 4 runs. Should have
        # 9 runs in total in the data store afterwards
        iactcopy = cscripts.csiactcopy()
        iactcopy['runlist']       = self._runlist2
        iactcopy['remote_master'] = self._remote
        iactcopy['prodname']      = 'unit-test'
        iactcopy['outpath']       = 'csiactcopy_py2'
        iactcopy['logfile']       = 'csiactcopy_py5.log'
        iactcopy['chatter']       = 4
        
        # Run csiactcopy script and save run list
        iactcopy.logFileOpen()   # Make sure we get a log file
        iactcopy.run()
        iactcopy.save()

        # Check copy
        self._check_copy('csiactcopy_py2')
        self._check_n_obs('csiactcopy_py2', 9)
        
        # Set-up csiactcopy using the same run list with 4 runs again, should
        # lead to skipping of files, and we should still have 9 runs at the
        # end in the datastore.
        iactcopy = cscripts.csiactcopy()
        iactcopy['runlist']       = self._runlist2
        iactcopy['remote_master'] = self._remote
        iactcopy['prodname']      = 'unit-test'
        iactcopy['outpath']       = 'csiactcopy_py2'
        iactcopy['logfile']       = 'csiactcopy_py6.log'
        iactcopy['chatter']       = 4
        iactcopy['clobber']       = True
        
        # Run csiactcopy script and save run list
        iactcopy.logFileOpen()   # Make sure we get a log file
        iactcopy.run()
        iactcopy.save()

        # Check copy
        self._check_copy('csiactcopy_py2')
        self._check_n_obs('csiactcopy_py2',9)

        # Return
        return

    # Check copy of IACT data store
    def _check_copy(self, pathname):
        """
        Check copy of IACT data store

        Parameters
        ----------
        pathname : str
            Path to copied IACT data store
        """
        # Set file names
        hdu_index_name = gammalib.GFilename(pathname+'/hdu-index.fits')
        obs_index_name = gammalib.GFilename(pathname+'/obs-index.fits')
        master_name    = gammalib.GFilename(pathname+'/master.json')

        # Check for existence of files
        self.test_assert(hdu_index_name.exists(),
                         'Check if file "hdu-index.fits" exists')
        self.test_assert(obs_index_name.exists(),
                         'Check if file "obs-index.fits" exists')
        self.test_assert(master_name.exists(),
                         'Check if file "master.json" exists')

        # Check if index files are FITS file
        self.test_assert(hdu_index_name.is_fits(),
                         'Check if file "hdu-index.fits" is FITS file')
        self.test_assert(obs_index_name.is_fits(),
                         'Check if file "obs-index.fits" is FITS file')
        
        # Return
        return

    # Check copy
    def _check_n_obs(self, pathname, n_expected):
        """
        Check number of available observations

        Parameters
        ----------
        pathname : str
            Path to copied IACT data store
        n_expected : int
            Expected number of observations in IACT data store
        """
        # Set file name
        obs_index_name = gammalib.GFilename(pathname+'/obs-index.fits[OBS_INDEX]')
        
        # Open index file
        fits = gammalib.GFits(obs_index_name)
        
        # Get number of observations
        n_obs = fits[obs_index_name.extname()].nrows()
        
        # Close FITS file
        fits.close()

        # Check for existence of observations
        self.test_value(n_obs, n_expected, 'Check for number of observations')
                
        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# Display summary of observation definition XML file
#
# Copyright (C) 2016 Michael Mayer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import gammalib
import cscripts
try:
    import matplotlib.pyplot as plt
    plt.figure()
    plt.close()
except (ImportError, RuntimeError):
    print('This script needs the "matplotlib" module')
    sys.exit()


# ==================== #
# Run csobsinfo script #
# ==================== #
def run_csobsinfo(filename, ra, dec, debug=True):
    """
    Run csobsinfo script

    Parameters
    ----------
    filename : str
        File name of observation definition XML file
    ra : str
        Target Right Ascension (deg)
    dec : str
        Target declination of pointing (deg)
    debug : bool, optional
        Switch on debugging in csobsinfo run

    Returns
    -------
    info : `~cscripts.csobsinfo`
        csobsinfo instance
    """
    # Get observation definition XML filename
    obsdef = gammalib.GFilename(filename)

    # Setup csobsinfo script
    info = cscripts.csobsinfo()
    info['inobs'] = obsdef.url()
    info['debug'] = debug

    # Set offset if the "ra" and "dec" arguments are not empty strings
    if ra != '' and dec != '':
        info['ra']     = float(ra)
        info['dec']    = float(dec)
        info['offset'] = True
    else:
        info['offset'] = False

    # Run csobsinfo
    info.run()

    # Return
    return info


# ================ #
# Plot information #
# ================ #
def plot_information(info, ra, dec, plotfile):
    """
    Plot information

    Parameters
    ----------
    info : `~cscripts.csobsinfo`
        csobsinfo instance
    ra : float
        Right Ascension (deg)
    dec : float
        Declination (deg)
    plotfile : str
        Plot filename
    """
    # Retrieve observation information
    zeniths  = info.zeniths()
    azimuths = info.azimuths()
    offsets  = info.offsets()
    ebounds  = info.ebounds()
    gti      = info.gti()

    # Create figure with subplots. In case that offset and energy boundary
    # information is available a column is added to the subplots
    nrows = 2
    ncols = 2
    if info['offset'].boolean() and ebounds.size() > 0:
        ncols += 1
    iplot = 1
    plt.figure()
    plt.subplot(nrows, ncols, iplot)

    # Plot zenith angle histogram
    zmin = min(zeniths)
    zmax = max(zeniths)
    plt.hist(zeniths, bins=30, range=(zmin, zmax), fc='blue')
    plt.xlabel('Zenith Angle (deg)')
    plt.ylabel('Frequency')
    plt.title('Zenith angle distribution')

    # Plot azimuth angle histogram
    iplot += 1
    plt.subplot(nrows, ncols, iplot)
    amin = min(azimuths)
    amax = max(azimuths)
    plt.hist(azimuths, bins=30, range=(amin, amax), fc='blue')
    plt.xlabel('Azimuth Angle (deg)')
    plt.ylabel('Frequency')
    plt.title('Azimuth distribution')

    # If available, plot offset from target histogram
    if info['offset'].boolean():
        iplot += 1
        plt.subplot(nrows, ncols, iplot)
        omin = min(offsets)
        omax = max(offsets)
        plt.hist(offsets, bins=30, range=(omin, omax), fc='blue')
        plt.xlabel('Offset from (RA, DEC)=('+str(ra)+','+str(dec)+') (deg)')
        plt.ylabel('Frequency')
        plt.title('Offset distribution')

    # If available, plot energy thresholds histogram
    if ebounds.size() > 0:
        emin = []
        emax = []
        for i in range(ebounds.size()):
            emin.append(ebounds.emin(i).log10TeV())
            emax.append(ebounds.emax(i).log10TeV())
        iplot += 1
        plt.subplot(nrows, ncols, iplot)
        plt.hist(emin, bins=80, range=(-1.0, 2.0), fc='red', label='emin')
        plt.hist(emax, bins=80, range=(-1.0, 2.0), fc='blue', label='emax')  
        plt.xlabel('Energy threshold (log10 (E/TeV))')
        plt.ylabel('Frequency')
        plt.legend(loc='upper left')
        plt.title('Energy threshold')

    # Plot zenith angle versus mean observing time for all observations
    iplot += 1
    plt.subplot(nrows, ncols, iplot)
    times = []
    for i in range(gti.size()):
        tmean = gti.tstart(i) + 0.5*(gti.tstart(i)-gti.tstart(i))
        times.append(tmean.mjd())
    plt.plot(times, zeniths, 'o', lw=2.0, color='black')
    plt.xlabel('Time (MJD)')
    plt.ylabel('Zenith Angle (deg)')
    plt.title('Observation time')

    # Show plots or save it into file
    if len(plotfile) > 0:
        plt.savefig(plotfile)
    else:
        plt.show()

    # Return
    return


# ================ #
# Show observation #
# ================ #
def show_obs():
    """
    Show observation
    """
    # Set usage string
    usage = 'show_obs.py [-ra ra] [-dec dec] [-p plotfile] file'

    # Set default options
    options = [{'option': '-ra',  'value': ''},
               {'option': '-dec', 'value': ''},
               {'option': '-p',   'value': ''}]

    # Get arguments and options from command line arguments
    args, options = cscripts.ioutils.get_args_options(options, usage)

    # Extract script parameters from options
    ra       = options[0]['value']
    dec      = options[1]['value']
    plotfile = options[2]['value']

    # Run csobsinfo
    info = run_csobsinfo(args[0], ra, dec)

    # Plot information
    plot_information(info, ra, dec, plotfile)

    # Return
    return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Show observation
    show_obs()



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the ctbutterfly tool.
#
# Copyright (C) 2014-2018 Michal Mayer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import gammalib
import ctools
from testing import test


# =============================== #
# Test class for ctbutterfly tool #
# =============================== #
class Test(test):
    """
    Test class for ctbutterfly tool
    """
    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('ctbutterfly')

        # Append tests
        self.append(self._test_cmd, 'Test ctbutterfly on command line')
        self.append(self._test_python, 'Test ctbutterfly from Python')

        # Return
        return

    # Test ctbutterfly on command line
    def _test_cmd(self):
        """
        Test ctbutterfly on the command line
        """
        # Set tool name
        ctbutterfly = self._tool('ctbutterfly')

        # Setup ctbutterfly command
        cmd = ctbutterfly+' inobs="'+self._events+'"'+\
                          ' outfile="ctbutterfly_cmd1.dat"'+ \
                          ' inmodel="'+self._model +'" srcname="Crab"'+ \
                          ' caldb="'+self._caldb+'" irf="'+self._irf+'"'+ \
                          ' emin=0.1 emax=100.0'+ \
                          ' logfile="ctbutterfly_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check result file
        self._check_result_file('ctbutterfly_cmd1.dat')

        # Setup ctbutterfly command
        cmd = ctbutterfly+' inobs="event_file_that_does_not_exist.fits"'+\
                          ' outfile="ctbutterfly_cmd2.dat"'+ \
                          ' inmodel="'+self._model +'" srcname="Crab"'+ \
                          ' caldb="'+self._caldb+'" irf="'+self._irf+'"'+ \
                          ' emin=0.1 emax=100.0'+ \
                          ' logfile="ctbutterfly_cmd2.log" debug=yes chatter=1'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check ctbutterfly --help
        self._check_help(ctbutterfly)

        # Return
        return

    # Test ctbutterfly from Python
    def _test_python(self):
        """
        Test ctbutterfly from Python
        """
        # Set-up ctbutterfly
        butterfly = ctools.ctbutterfly()
        butterfly['inobs']   = self._events
        butterfly['inmodel'] = self._model
        butterfly['srcname'] = 'Crab'
        butterfly['caldb']   = self._caldb
        butterfly['irf']     = self._irf
        butterfly['emin']    = 0.1
        butterfly['emax']    = 100.0
        butterfly['outfile'] = 'ctbutterfly_py1.dat'
        butterfly['logfile'] = 'ctbutterfly_py1.log'
        butterfly['chatter'] = 2

        # Run ctbutterfly tool
        butterfly.logFileOpen()   # Make sure we get a log file
        butterfly.run()
        butterfly.save()

        # Check result file
        self._check_result_file('ctbutterfly_py1.dat')

        # Check CSV result
        self.test_value(butterfly.butterfly().nrows(), 100,
             'Check for 100 rows in CSV file')
        self.test_value(butterfly.butterfly().ncols(), 4,
             'Check for 4 columns in CSV file')

        # Set-up ctbutterfly
        butterfly = ctools.ctbutterfly()
        butterfly['inobs']   = self._events
        butterfly['inmodel'] = self._model
        butterfly['srcname'] = 'Crab'
        butterfly['caldb']   = self._caldb
        butterfly['irf']     = self._irf
        butterfly['method']  = 'ENVELOPE'
        butterfly['emin']    = 0.1
        butterfly['emax']    = 100.0
        butterfly['fit']     = True
        butterfly['outfile'] = 'ctbutterfly_py2.dat'
        butterfly['logfile'] = 'ctbutterfly_py2.log'
        butterfly['chatter'] = 3

        # Execute ctbutterfly tool
        butterfly.logFileOpen()   # Make sure we get a log file
        butterfly.execute()

        # Check result file
        self._check_result_file('ctbutterfly_py2.dat')

        # Recover observation container
        #obs = butterfly.obs()

        # TODO: Do someting test on observation container

        # Copy ctbkgcube tool and execute copy
        cpy_butterfly = butterfly
        cpy_butterfly['outfile'] = 'ctbutterfly_py3.dat'
        cpy_butterfly['logfile'] = 'ctbutterfly_py3.log'
        cpy_butterfly['chatter'] = 4

        # Execute ctbutterfly tool
        cpy_butterfly.logFileOpen()   # Make sure we get a log file
        cpy_butterfly.execute()

        # Check result file
        self._check_result_file('ctbutterfly_py3.dat')

        # Clear ctbkgcube tool
        butterfly.clear()

        # TODO: Do some test after clearing

        # Return
        return

    # Check result file
    def _check_result_file(self, filename):
        """
        Check result file
        """
        # Open result file as CSV file
        results = gammalib.GCsv(filename)

        # Check dimensions
        self.test_value(results.nrows(), 100,
             'Check for 100 rows in butterfly file')
        self.test_value(results.ncols(), 4,
             'Check for 4 columns in butterfly file')

        # Return
        return



#!/usr/bin/env python
# ==========================================================================
# Create commands to split TS map computation
#
# Copyright (C) 2016-2019 Michael Mayer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import os
import sys
import math
import gammalib
import ctools


# ================== #
# cstsmapsplit class #
# ================== #
class cstsmapsplit(ctools.cslikelihood):
    """
    Generates commands to split TS map computation

    This class implements the creation of file containing commands to split
    the computation of a TS map into several machines. It derives from
    the ctools.cscript class which provides support for parameter files,
    command line arguments, and logging. In that way the Python script
    behaves just as a regular ctool. 
    """

    # Consructor
    def __init__(self, *argv):
        """
        Constructor
        """
        # Initialise application by calling the appropriate class constructor
        self._init_cslikelihood(self.__class__.__name__, ctools.__version__, argv)

        # Set data members
        self._outmap       = gammalib.GFilename()
        self._bins_per_job = 0
        self._compute_null = False
        self._outfile      = gammalib.GFilename()
        self._map          = gammalib.GSkyMap()
        self._cmd          = []
        self._srcname      = ''

        # Return
        return


    # Private methods
    def _get_parameters(self):
        """
        Get parameters from parfile and setup the observation
        """
        # If there are no observations in container then get some ...
        if self.obs().size() == 0:
            self.obs(self._get_observations())

        # Set models if we have none
        if self.obs().models().size() == 0:
            self.obs().models(self['inmodel'].filename())

        # Get TS map paramters
        self._map     = self._create_map(self.obs())
        self._srcname = self['srcname'].string()
        self._outmap  = self['outmap'].filename()

        # Get additional parameters
        self._bins_per_job = self['bins_per_job'].integer()
        self._compute_null = self['compute_null'].boolean()
        self._outfile      = self['outfile'].filename()

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)

        # Return
        return

    def _compute_null_hypothesis(self):
        """
        Compute null hypothesis

        Returns
        -------
        logl : float
            Log-likelihood of null hypothesis
        """
        # Store original models  
        models_orig = self.obs().models()

        # Get models instance
        models0 = self.obs().models()

        # Remove test source   
        models0.remove(self._srcname)

        # Set models for null hypothesis 
        self.obs().models(models0)

        # Optimise observation container
        self.obs().optimize(self.opt())

        # Set original models to container
        self.obs().models(models_orig)

        # Return optimised log-likelihood of null hypothesis
        return -(self.opt().value())


    # Public methods
    def run(self):
        """
        Run the script
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()

        # Write information into logger
        self._log_header1(gammalib.TERSE, 'Test source')
        self._log_string(gammalib.TERSE, str(self.obs().models()[self._srcname]))

        # Set log-likelihood to zero
        logL0 = 0.0

        # Pre-compute null hypothesis if requested
        if self._compute_null:

            # Write information into logger
            self._log_header1(gammalib.TERSE, 'Compute null hypothesis')

            # Compute null hypothesis
            logL0 = self._compute_null_hypothesis()

            # Write likelihood into logger
            self._log_value(gammalib.TERSE, 'Source removed', self._srcname)
            self._log_value(gammalib.TERSE, 'Log-likelihood', repr(logL0))

        # Get parameters of TS map ctool
        pars = gammalib.GApplicationPars('cttsmap.par')

        # Compute total number of jobs
        nbins = self._map.npix()
        njobs = int(math.ceil(float(nbins) / float(self._bins_per_job)) + 0.1)   

        # Set parameters to be skipped now, we will deal with them later
        skip_pars = ['binmin', 'binmax', 'logL0', 'outmap', 'logfile']

        # Set tool name for computation
        base_command = 'cttsmap'

        # Write information into logger
        self._log_header1(gammalib.TERSE, 'Create commands')
        self._log_value(gammalib.TERSE, 'Number of cttsmap calls', njobs)

        # Loop over TS map parameters
        for par in pars:

            # Skip if we deal with them later
            if par.name() in skip_pars:
                continue

            # Skip if parameter was not queried
            if not self[par.name()].was_queried():
                continue

            # Set TS map parameter according to input from
            # this script
            par.value(self[par.name()].value())

            # Append command to set parameter
            base_command += ' ' + par.name() + '=' + par.value()

        # Append null hypothesis parameter
        base_command += ' logL0=' + repr(logL0)  

        # Set binning to start from zero
        binmin = 0
        binmax = 0

        # Clear command sequence
        self._cmd = []

        # Loop over jobs and create commands
        for job in range(njobs):

            # Set specific outmap file name
            outmap = self._outmap.url().replace('.fits', '_'+str(job)+'.fits')

            # Set bin numbers to be computed in this job
            binmin = binmax
            binmax = binmin + self._bins_per_job
            if binmax > nbins:
                binmax = nbins

            # Setup sliced command
            sliced_command = '%s binmin=%s binmax=%s outmap=%s logfile=%s' % \
                             (base_command, str(binmin), str(binmax),
                              outmap, outmap.replace('.fits','.log'))

            # If running in background is requested then append a &
            if self['run_in_bkg'].boolean():
                sliced_command += ' &'

            # Append command to list of commands
            self._cmd.append(sliced_command)

        # Write information into logger
        if self._logExplicit():
            self._log('\n')
            self._log.header2('Commands')
            for cmd in self._cmd:
                self._log(cmd)
                self._log('\n')
            self._log('\n')

        # Return
        return

    def save(self):
        """
        Save commands to ASCII file
        """
        # Get filename
        filename = self._outfile.url()

        # Open file
        f = open(filename, 'w')

        # Write commands to file
        for cmd in self._cmd:
            f.write(cmd + '\n')

        # Close file
        f.close()

        # Make file executable
        os.system('chmod +x %s' % filename)

        # Return
        return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = cstsmapsplit(sys.argv)

    # Execute application
    app.execute()



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the csinfo script.
#
# Copyright (C) 2016-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
from testing import test


# ============================ #
# Test class for csinfo script #
# ============================ #
class Test(test):
    """
    Test class for csinfo script
    """

    # Constructor
    def __init__(self):
        """
        Constructor.
        """
        # Call base class constructor
        test.__init__(self)

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions.
        """
        # Set test name
        self.name('csinfo')

        # Append tests
        self.append(self._test_cmd, 'Test csinfo on command line')

        # Return
        return

    # Test csinfo on command line
    def _test_cmd(self):
        """
        Test csinfo on the command line.
        """
        # Set script name
        csinfo = self._script('csinfo')

        # Setup csinfo command
        cmd = csinfo

        # Execute script without arguments
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Execute script with "list" argument
        self.test_assert(self._execute(cmd+' list') == 0,
             'Check successful execution with "list" argument from command line')

        # Execute script with "check" argument
        self.test_assert(self._execute(cmd+' check') == 0,
             'Check successful execution with "check" argument from command line')

        # Execute script with "info" argument
        self.test_assert(self._execute(cmd+' info') == 0,
             'Check successful execution with "info" argument from command line')

        # Execute script with invalid argument
        self.test_assert(self._execute(cmd+' not_a_valid_argument',
                         success=False) != 0,
                         'Check unsuccessful execution with unsupported '
                         'argument from command line')

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the cslightcrv script.
#
# Copyright (C) 2016-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import gammalib
import cscripts
from testing import test


# ============================= #
# Test class for csebins script #
# ============================= #
class Test(test):
    """
    Test class for csebins script

    This test class makes unit tests for the csebins script by using it from
    the command line and from Python.
    """
    
    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions.
        """
        # Set test name
        self.name('csebins')

        # Append tests
        self.append(self._test_cmd, 'Test csebins on command line')
        self.append(self._test_python, 'Test csebins from Python')

        # Return
        return

    # Test csebins on command line
    def _test_cmd(self):
        """
        Test csebins on the command line.
        """
        # Set script name
        csebins = self._script('csebins')

        # Setup csebins command
        cmd = csebins+' inobs="NONE"'+ \
                      ' caldb="'+self._caldb+'" irf="'+self._irf+'"'+ \
                      ' emin=0.1 emax=100.0'+ \
                      ' outfile="csebins_cmd1.fits"'+ \
                      ' logfile="csebins_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check energy boundaries
        self._check_ebounds('csebins_cmd1.fits', 33)

        # Setup csebins command
        cmd = csebins+' inobs="NONE"'+ \
                      ' caldb="Database_that_does_not_exist" irf="'+self._irf+'"'+ \
                      ' emin=0.1 emax=100.0'+ \
                      ' outfile="csebins_cmd2.fits"'+ \
                      ' logfile="csebins_cmd2.log" debug=yes chatter=1'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check csebins --help
        self._check_help(csebins)

        # Return
        return

    # Test csebins from Python
    def _test_python(self):
        """
        Test csebins from Python
        """
        # Set-up csebins
        ebins = cscripts.csebins()
        ebins['inobs']   = 'NONE'
        ebins['caldb']   = self._caldb
        ebins['irf']     = self._irf
        ebins['emin']    = 0.1
        ebins['emax']    = 100.0
        ebins['outfile'] = 'csebins_py1.fits'
        ebins['logfile'] = 'csebins_py1.log'
        ebins['chatter'] = 2

        # Run csebins script and energy boundaries
        ebins.logFileOpen()   # Make sure we get a log file
        ebins.run()
        ebins.save()

        # Check energy boundaries
        self._check_ebounds('csebins_py1.fits', 33)

        # Return
        return

    # Check energy boundaries result file
    def _check_ebounds(self, filename, bins):
        """
        Check energy boundaries file
        """
        # Load energy boundaries file
        ebounds = gammalib.GEbounds(filename)

        # Check number of energy boundaries
        self.test_value(ebounds.size(), bins)
        
        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# This script generates the pull distribution for all model parameters.
#
# Copyright (C) 2011-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import gammalib
import ctools
from cscripts import obsutils
from cscripts import ioutils
from cscripts import mputils


# ============ #
# cspull class #
# ============ #
class cspull(ctools.csobservation):
    """
    Generates pull distributions for a model
    """
    # Constructor
    def __init__(self, *argv):
        """
        Constructor
        """
        # Initialise application by calling the appropriate class constructor
        self._init_csobservation(self.__class__.__name__, ctools.__version__, argv)

        # Initialise some members
        self._nthreads = 0

        # Return
        return

    # State methods por pickling
    def __getstate__(self):
        """
        Extend ctools.csobservation getstate method to include some members

        Returns
        -------
        state : dict
            Pickled instance
        """
        # Set pickled dictionary
        state = {'base'     : ctools.csobservation.__getstate__(self),
                 'nthreads' : self._nthreads}

        # Return pickled dictionary
        return state

    def __setstate__(self, state):
        """
        Extend ctools.csobservation setstate method to include some members

        Parameters
        ----------
        state : dict
            Pickled instance
        """
        ctools.csobservation.__setstate__(self, state['base'])
        self._nthreads = state['nthreads']

        # Return
        return

    # Private methods
    def _get_parameters(self):
        """
        Get parameters from parfile
        """
        # If there are no observations in container then get some ...
        if self.obs().is_empty():
            self.obs(self._get_observations())

        # ... otherwise add response information and energy boundaries
        # in case they are missing
        else:
            self._setup_observations(self.obs())

        # Set observation statistic
        self._set_obs_statistic(gammalib.toupper(self['statistic'].string()))

        # Get number of energy bins
        enumbins = self['enumbins'].integer()

        # Query parameters for On/Off observation
        if gammalib.toupper(self['onsrc'].string()) != 'NONE':
            self['onrad'].real()

        # Query parameters for binned if requested
        elif enumbins > 0:
            self['npix'].integer()
            self['binsz'].real()
            self['coordsys'].string()
            self['proj'].string()

        # Set models if we have none
        if self.obs().models().is_empty():
            self.obs().models(self['inmodel'].filename())

        # Query other parameters
        self['ntrials'].integer()
        self['edisp'].boolean()
        self['seed'].integer()
        self['chatter'].integer()

        # Query some parameters
        self['outfile'].filename()
        self['profile'].boolean()

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)

        # Set number of processes for multiprocessing
        self._nthreads = mputils.nthreads(self)

        # Return
        return

    def _obs_string(self, obs):
        """
        Generate summary string for observation

        Parameters
        ----------
        obs : `~gammalib.GCTAObservation`
            Observation

        Returns
        -------
        text : str
            Summary string
        """
        # Extract information from observation
        if obs.classname() == 'GCTAOnOffObservation':
            emin = obs.on_spec().ebounds().emin().TeV()
            emax = obs.on_spec().ebounds().emax().TeV()
            mode = 'On/Off'
        else:
            emin = obs.events().ebounds().emin().TeV()
            emax = obs.events().ebounds().emax().TeV()
            binned = (obs.events().classname() == 'GCTAEventCube')
            if binned:
                mode = 'binned'
            else:
                mode = 'unbinned'
        events = obs.nobserved()

        # Compose summary string
        if events > 0:
            text = '%d events, %.3f-%.3f TeV, %s' % (events, emin, emax, mode)
        else:
            text = '%.3f-%.3f TeV, %s' % (emin, emax, mode)

        # Return summary string
        return text

    def _trial(self, seed):
        """
        Compute the pull for a single trial

        Parameters
        ----------
        seed : int
            Random number generator seed

        Returns
        -------
        result : dict
            Dictionary of results
        """
        # Write header
        self._log_header2(gammalib.NORMAL, 'Trial %d' %
                          (seed-self['seed'].integer()+1))

        # Get number of energy bins and On source name and initialise
        # some parameters
        nbins     = self['enumbins'].integer()
        onsrc     = self['onsrc'].string()
        edisp     = self['edisp'].boolean()
        statistic = self['statistic'].string()
        emin      = None
        emax      = None
        binsz     = 0.0
        npix      = 0
        proj      = 'TAN'
        coordsys  = 'CEL'

        # If we have a On source name then set On region radius
        if gammalib.toupper(onsrc) != 'NONE':
            onrad = self['onrad'].real()
            emin  = self['emin'].real()
            emax  = self['emax'].real()
            edisp = True   # Use always energy dispersion for On/Off
        else:

            # Reset On region source name and radius
            onrad = 0.0
            onsrc = None

            # If we have a binned obeservation then specify the lower and
            # upper energy limit in TeV
            if nbins > 0:
                emin     = self['emin'].real()
                emax     = self['emax'].real()
                binsz    = self['binsz'].real()
                npix     = self['npix'].integer()
                proj     = self['proj'].string()
                coordsys = self['coordsys'].string()

        # Simulate events
        obs = obsutils.sim(self.obs(),
                           emin=emin, emax=emax, nbins=nbins,
                           onsrc=onsrc, onrad=onrad,
                           addbounds=True, seed=seed,
                           binsz=binsz, npix=npix, proj=proj, coord=coordsys,
                           edisp=edisp, nthreads=1, log=False,
                           debug=self._logDebug(), chatter=self['chatter'].integer())

        # Determine number of events in simulation
        nevents = 0.0
        for run in obs:
            nevents += run.nobserved()

        # Write simulation results
        self._log_header3(gammalib.NORMAL, 'Simulation')
        for run in self.obs():
            self._log_value(gammalib.NORMAL, 'Input observation %s' % run.id(),
                            self._obs_string(run))
        for run in obs:
            self._log_value(gammalib.NORMAL, 'Output observation %s' % run.id(),
                            self._obs_string(run))
        self._log_value(gammalib.NORMAL, 'Number of simulated events', nevents)

        # Fit model
        if self['profile'].boolean():
            models = self.obs().models()
            for model in models:
                like = ctools.cterror(obs)
                like['srcname']   = model.name()
                like['edisp']     = edisp
                like['statistic'] = statistic
                like['nthreads']  = 1  # Avoids OpenMP conflict
                like['debug']     = self._logDebug()
                like['chatter']   = self['chatter'].integer()
                like.run()
        else:
            like = ctools.ctlike(obs)
            like['edisp']     = edisp
            like['statistic'] = statistic
            like['nthreads']  = 1  # Avoids OpenMP conflict
            like['debug']     = self._logDebug()
            like['chatter']   = self['chatter'].integer()
            like.run()

        # Store results
        logL   = like.opt().value()
        npred  = like.obs().npred()
        models = like.obs().models()

        # Write result header
        self._log_header3(gammalib.NORMAL, 'Pulls')

        # Gather results in form of a list of result columns and a
        # dictionary containing the results. The result contains the
        # log-likelihood, the number of simulated events, the number of
        # predicted events and for each fitted parameter the fitted value,
        # the pull and the fit error.
        #
        # Note that we do not use the model and parameter iterators
        # because we need the indices to get the true (or real) parameter
        # values from the input models.
        colnames = ['LogL', 'Sim_Events', 'Npred_Events']
        values   = {'LogL': logL, 'Sim_Events': nevents, 'Npred_Events': npred}
        for i in range(models.size()):
            model = models[i]
            for k in range(model.size()):
                par = model[k]
                if par.is_free():

                    # Set name as a combination of model name and parameter
                    # name separated by an underscore. In that way each
                    # parameter has a unique name.
                    name = model.name()+'_'+par.name()

                    # Append parameter, Pull_parameter and e_parameter column
                    # names
                    colnames.append(name)
                    colnames.append('Pull_'+name)
                    colnames.append('e_'+name)

                    # Compute pull for this parameter as the difference
                    #               (fitted - true) / error
                    # In case that the error is 0 the pull is set to 99
                    fitted_value = par.value()
                    real_value   = self.obs().models()[i][k].value()
                    error        = par.error()
                    if error != 0.0:
                        pull = (fitted_value - real_value) / error
                    else:
                        pull = 99.0

                    # Store results in dictionary
                    values[name]         = fitted_value
                    values['Pull_'+name] = pull
                    values['e_'+name]    = error

                    # Write results into logger
                    value = '%.4f (%e +/- %e)' % (pull, fitted_value, error)
                    self._log_value(gammalib.NORMAL, name, value)

        # Bundle together results in a dictionary
        result = {'colnames': colnames, 'values': values}

        # Return
        return result


    # Public methods
    def run(self):
        """
        Run the script
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()

        # Write observation into logger
        self._log_observations(gammalib.NORMAL, self.obs(), 'Input observation')

        # Write header
        self._log_header1(gammalib.TERSE, 'Generate pull distribution')

        # Get number of trials
        ntrials = self['ntrials'].integer()

        # Get seed value
        seed = self['seed'].integer()

        # If more than a single thread is requested then use multiprocessing
        if self._nthreads > 1:
            args        = [(self, '_trial', i + seed) for i in range(ntrials)]
            poolresults = mputils.process(self._nthreads, mputils.mpfunc, args)

        # Continue with regular processing
        for i in range(ntrials):

            # If multiprocessing was used then recover results and put them
            # into the log file
            if self._nthreads > 1:
                result = poolresults[i][0]
                self._log_string(gammalib.TERSE, poolresults[i][1]['log'], False)

            # ... otherwise make a trial now
            else:
                result = self._trial(i + seed)

            # Write out trial result
            ioutils.write_csv_row(self['outfile'].filename().url(), i,
                                  result['colnames'], result['values'])

        # Return
        return

    def models(self, models):
        """
        Set model

        Parameters
        ----------
        models : `~gammalib.GModels`
            Set model container
        """
        # Set model container
        self.obs().models(models)

        # Return
        return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = cspull(sys.argv)

    # Execute application
    app.execute()



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the cslightcrv script.
#
# Copyright (C) 2016-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import gammalib
import cscripts
from testing import test


# ================================ #
# Test class for cslightcrv script #
# ================================ #
class Test(test):
    """
    Test class for cslightcrv script

    This test class makes unit tests for the cslightcrv script by using it
    from the command line and from Python.
    """

    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Add off-axis events for classical analysis
        self._offaxis_events = self._datadir + '/crab_offaxis1.fits'
        self._model_onoff    = self._datadir + '/crab_onoff.xml'

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions.
        """
        # Set test name
        self.name('cslightcrv')

        # Append tests
        self.append(self._test_cmd, 'Test cslightcrv on command line')
        self.append(self._test_python, 'Test cslightcrv from Python')
        self.append(self._test_pickeling, 'Test cslightcrv pickeling')

        # Return
        return

    # Test cslightcrv on command line
    def _test_cmd(self):
        """
        Test cslightcrv on the command line.
        """
        # Set script name
        cslightcrv = self._script('cslightcrv')

        # Setup cslightcrv command
        cmd = cslightcrv+' inobs="'+self._events+'"'+ \
                         ' inmodel="'+self._model+'" srcname="Crab"'+ \
                         ' caldb="'+self._caldb+'" irf="'+self._irf+'"'+ \
                         ' tbinalg=LIN '+ \
                         ' tmin=2020-01-01T00:00:00 tmax=2020-01-01T00:05:00'+ \
                         ' tbins=2 method=3D enumbins=0 emin=1.0 emax=100.0'+ \
                         ' outfile="lightcurve_cmd1.fits"'+ \
                         ' logfile="cslightcrv_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check light curve
        self._check_light_curve('lightcurve_cmd1.fits', 2)

        # Setup cslightcrv command
        cmd = cslightcrv+' inobs="events_that_do_not_exist.fits"'+ \
                         ' inmodel="'+self._model+'" srcname="Crab"'+ \
                         ' caldb="'+self._caldb+'" irf="'+self._irf+'"'+ \
                         ' tbinalg=LIN '+ \
                         ' tmin=2020-01-01T00:00:00 tmax=2020-01-01T00:05:00'+ \
                         ' tbins=2 method=3D enumbins=0 emin=1.0 emax=100.0'+ \
                         ' outfile="lightcurve_cmd1.fits"'+ \
                         ' logfile="cslightcrv_cmd2.log" debug=yes'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check cslightcrv --help
        self._check_help(cslightcrv)

        # Return
        return

    # Test cslightcrv from Python
    def _test_python(self):
        """
        Test cslightcrv from Python
        """
        # Set-up unbinned cslightcrv
        lcrv = cscripts.cslightcrv()
        lcrv['inobs']    = self._events
        lcrv['inmodel']  = self._model
        lcrv['srcname']  = 'Crab'
        lcrv['caldb']    = self._caldb
        lcrv['irf']      = self._irf
        lcrv['tbinalg']  = 'LIN'
        lcrv['tmin']     = '2020-01-01T00:00:00'
        lcrv['tmax']     = '2020-01-01T00:05:00'
        lcrv['tbins']    = 2
        lcrv['method']   = '3D'
        lcrv['enumbins'] = 0
        lcrv['emin']     = 1.0
        lcrv['emax']     = 100.0
        lcrv['outfile']  = 'cslightcrv_py1.fits'
        lcrv['logfile']  = 'cslightcrv_py1.log'
        lcrv['chatter']  = 2
        lcrv['publish']  = True

        # Run cslightcrv script and save light curve
        lcrv.logFileOpen()   # Make sure we get a log file
        lcrv.run()
        lcrv.save()

        # Check light curve
        self._check_light_curve('cslightcrv_py1.fits', 2)

        # Now use FILE as time bin algorithm. For this we need first to
        # create an ASCII file. We use now 2 time bins. The ASCII file
        # is saved into the file "lightcurve_py2.dat".
        csv    = gammalib.GCsv(2,2)
        tmin   = 58849.00
        tdelta = 0.0017361
        for i in range(csv.nrows()):
            csv[i,0] = '%.5f' % (tmin +  i    * tdelta)
            csv[i,1] = '%.5f' % (tmin + (i+1) * tdelta)
        csv.save('cslightcrv_py2.dat', ' ', True)

        # Set-up unbinned cslightcrv
        lcrv = cscripts.cslightcrv()
        lcrv['inobs']    = self._events
        lcrv['inmodel']  = self._model
        lcrv['srcname']  = 'Crab'
        lcrv['caldb']    = self._caldb
        lcrv['irf']      = self._irf
        lcrv['tbinalg']  = 'FILE'
        lcrv['tbinfile'] = 'cslightcrv_py2.dat'
        lcrv['method']   = '3D'
        lcrv['enumbins'] = 0
        lcrv['emin']     = 1.0
        lcrv['emax']     = 100.0
        lcrv['fix_bkg']  =  True
        lcrv['outfile']  = 'cslightcrv_py2.fits'
        lcrv['logfile']  = 'cslightcrv_py2.log'
        lcrv['chatter']  = 3

        # Execute cslightcrv script
        lcrv.execute()

        # Check light curve
        self._check_light_curve('cslightcrv_py2.fits', 2)

        # Now we setup an observation container on input. We attached the
        # model to the observation container so that cslightcrv should
        # no longer query for the parameter.
        cta = gammalib.GCTAObservation(self._events)
        obs = gammalib.GObservations()
        obs.append(cta)
        obs.models(self._model)

        # Set-up unbinned cslightcrv from observation container. Now use
        # the GTI algorithm so that we test all timing algorithms.
        lcrv = cscripts.cslightcrv(obs)
        lcrv['srcname']  = 'Crab'
        lcrv['caldb']    = self._caldb
        lcrv['irf']      = self._irf
        lcrv['tbinalg']  = 'GTI'
        lcrv['method']   = '3D'
        lcrv['enumbins'] = 0
        lcrv['emin']     = 1.0
        lcrv['emax']     = 100.0
        lcrv['outfile']  = 'cslightcrv_py3.fits'
        lcrv['logfile']  = 'cslightcrv_py3.log'
        lcrv['chatter']  = 4

        # Execute cslightcrv script
        lcrv.execute()

        # Check light curve
        self._check_light_curve('cslightcrv_py3.fits', 1)

        # Binned cslightcrv
        lcrv = cscripts.cslightcrv()
        lcrv['inobs']    = self._events
        lcrv['inmodel']  = self._model
        lcrv['srcname']  = 'Crab'
        lcrv['caldb']    = self._caldb
        lcrv['irf']      = self._irf
        lcrv['tbinalg']  = 'LIN'
        lcrv['tmin']     = '2020-01-01T00:00:00'
        lcrv['tmax']     = '2020-01-01T00:05:00'
        lcrv['tbins']    = 2
        lcrv['method']   = '3D'
        lcrv['emin']     = 1.0
        lcrv['emax']     = 100.0
        lcrv['enumbins'] = 3
        lcrv['coordsys'] = 'CEL'
        lcrv['proj']     = 'TAN'
        lcrv['xref']     = 83.63
        lcrv['yref']     = 22.01
        lcrv['nxpix']    = 10
        lcrv['nypix']    = 10
        lcrv['binsz']    = 0.04
        lcrv['outfile']  = 'cslightcrv_py4.fits'
        lcrv['logfile']  = 'cslightcrv_py4.log'
        lcrv['chatter']  = 4

        # Execute cslightcrv script
        lcrv.execute()

        # Check light curve
        self._check_light_curve('cslightcrv_py4.fits', 2)

        # cslightcrv with classical analysis
        lcrv = cscripts.cslightcrv()
        lcrv['inobs']         = self._offaxis_events
        lcrv['inmodel']       = self._model_onoff
        lcrv['srcname']       = 'Crab'
        lcrv['caldb']         = self._caldb
        lcrv['irf']           = self._irf
        lcrv['tbinalg']       = 'LIN'
        lcrv['tmin']          = '2020-01-01T00:00:00'
        lcrv['tmax']          = '2020-01-01T00:05:00'
        lcrv['tbins']         = 2
        lcrv['method']        = 'ONOFF'
        lcrv['use_model_bkg'] = False  # Needed even if WSTAT
        lcrv['emin']          = 1.0
        lcrv['emax']          = 100.0
        lcrv['enumbins']      = 2
        lcrv['coordsys']      = 'CEL'
        lcrv['xref']          = 83.63
        lcrv['yref']          = 22.01
        lcrv['rad']           = 0.2
        lcrv['etruemin']      = 1.0
        lcrv['etruemax']      = 100.0
        lcrv['etruebins']     = 5
        lcrv['statistic']     = 'WSTAT'
        lcrv['outfile']       = 'cslightcrv_py5.fits'
        lcrv['logfile']       = 'cslightcrv_py5.log'
        lcrv['chatter']       = 4

        # Execute cslightcrv script
        lcrv.execute()

        # Check light curve
        self._check_light_curve('cslightcrv_py5.fits', 2)

        # Set-up cslightcrv without multiprocessing
        lcrv = cscripts.cslightcrv()
        lcrv['inobs']    = self._events
        lcrv['inmodel']  = self._model
        lcrv['srcname']  = 'Crab'
        lcrv['caldb']    = self._caldb
        lcrv['irf']      = self._irf
        lcrv['tbinalg']  = 'LIN'
        lcrv['tmin']     = '2020-01-01T00:00:00'
        lcrv['tmax']     = '2020-01-01T00:05:00'
        lcrv['tbins']    = 2
        lcrv['method']   = '3D'
        lcrv['enumbins'] = 0
        lcrv['emin']     = 1.0
        lcrv['emax']     = 100.0
        lcrv['outfile']  = 'cslightcrv_py6.fits'
        lcrv['logfile']  = 'cslightcrv_py6.log'
        lcrv['chatter']  = 2
        lcrv['publish']  = True
        lcrv['nthreads'] = 1

        # Run cslightcrv script and save light curve
        lcrv.logFileOpen()  # Make sure we get a log file
        lcrv.run()
        lcrv.save()

        # Check light curve
        self._check_light_curve('cslightcrv_py6.fits', 2)

        # Return
        return

    # Test cslightcrv pickeling
    def _test_pickeling(self):
        """
        Test cslightcrv pickeling
        """
        # Perform pickeling tests of empty class
        self._pickeling(cscripts.cslightcrv())

        # Set-up unbinned cslightcrv
        lcrv = cscripts.cslightcrv()
        lcrv['inobs']    = self._events
        lcrv['inmodel']  = self._model
        lcrv['srcname']  = 'Crab'
        lcrv['caldb']    = self._caldb
        lcrv['irf']      = self._irf
        lcrv['tbinalg']  = 'LIN'
        lcrv['tmin']     = '2020-01-01T00:00:00'
        lcrv['tmax']     = '2020-01-01T00:05:00'
        lcrv['tbins']    = 2
        lcrv['method']   = '3D'
        lcrv['enumbins'] = 0
        lcrv['emin']     = 1.0
        lcrv['emax']     = 100.0
        lcrv['outfile']  = 'cslightcrv_py1_pickle.fits'
        lcrv['logfile']  = 'cslightcrv_py1_pickle.log'
        lcrv['chatter']  = 2
        lcrv['publish']  = True

        # Perform pickeling tests of filled class
        obj = self._pickeling(lcrv)

        # Run cslightcrv script and save light curve
        obj.logFileOpen()   # Make sure we get a log file
        obj.run()
        obj.save()

        # Check light curve
        self._check_light_curve('cslightcrv_py1_pickle.fits', 2)

        # Return
        return

    # Check light curve result file
    def _check_light_curve(self, filename, bins, prefactor=5.7e-16):
        """
        Check light curve file
        """
        # Expected column names
        cols = ['MJD', 'e_MJD', 'Prefactor', 'e_Prefactor',
                'Index', 'e_Index', 'TS', 'DiffUpperLimit',
                'FluxUpperLimit', 'EFluxUpperLimit']

        # Open FITS file
        fits = gammalib.GFits(filename)

        # Check FITS file structure
        self.test_value(fits.size(), 2,
             'Check for 2 extensions in light curve FITS file')
        self.test_assert(fits.contains('LIGHTCURVE'),
             'FITS file contains "LIGHTCURVE" extension')

        # Get LIGHTCURVE table
        table = fits['LIGHTCURVE']

        # Check FITS table structure
        self.test_value(table.ncols(), len(cols),
             'Check for %d columns in light curve FITS table' % len(cols))
        self.test_value(table.nrows(), bins,
             'Check for %d rows in light curve FITS table' % bins)
        for col in cols:
            self.test_assert(table.contains(col),
                 'FITS file contains "'+col+'" column')

        # Check that the Prefactor has the right order of magnitude
        for s in range(table.nrows()):
            self.test_value(table['Prefactor'][s], prefactor, 0.9*prefactor,
                            'Check prefactor value')

        # Close FITS file
        fits.close()

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# Generates a spectrum.
#
# Copyright (C) 2014-2019 Michael Mayer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import gammalib
import ctools
from cscripts import mputils


# ============ #
# csspec class #
# ============ #
class csspec(ctools.csobservation):
    """
    Generates a spectrum

    This class implements the generation of a Spectral Energy Distribution
    (SED) from gamma-ray observations.
    """

    # Constructors and destructors
    def __init__(self, *argv):
        """
        Constructor
        """
        # Initialise application by calling the appropriate class constructor
        self._init_csobservation(self.__class__.__name__, ctools.__version__, argv)

        # Initialise data members
        self._ebounds     = gammalib.GEbounds()
        self._fits        = None
        self._binned_mode = False
        self._onoff_mode  = False
        self._method      = 'AUTO'
        self._nthreads    = 0

        # Return
        return

    def __del__(self):
        """
        Destructor
        """
        # Return
        return

    # State methods for pickling
    def __getstate__(self):
        """
        Extend ctools.csobservation getstate method to include some members

        Returns
        -------
        state : dict
            Pickled instance
        """
        # Set pickled dictionary
        state = {'base'         : ctools.csobservation.__getstate__(self),
                 'ebounds'      : self._ebounds,
                 'fits'         : self._fits,
                 'binned_mode'  : self._binned_mode,
                 'onoff_mode'   : self._onoff_mode,
                 'method'       : self._method,
                 'nthreads'     : self._nthreads}

        # Return pickled dictionary
        return state

    def __setstate__(self, state):
        """
        Extend ctools.csobservation setstate method to include some members

        Parameters
        ----------
        state : dict
            Pickled instance
        """
        ctools.csobservation.__setstate__(self, state['base'])
        self._ebounds       = state['ebounds']
        self._fits          = state['fits']
        self._binned_mode   = state['binned_mode']
        self._onoff_mode    = state['onoff_mode']
        self._method        = state['method']
        self._nthreads      = state['nthreads']

        # Return
        return


    # Private methods
    def _get_parameters(self):
        """
        Get parameters from parfile and setup the observation
        """
        # Set observation if not done before
        if self.obs().is_empty():
            self._require_inobs('csspec::get_parameters()')
            self.obs(self._get_observations())

        # Set observation statistic
        self._set_obs_statistic(gammalib.toupper(self['statistic'].string()))

        # Set models if we have none
        if self.obs().models().is_empty():
            self.obs().models(self['inmodel'].filename())

        # Query source name
        self['srcname'].string()

        # Get spectrum generation method
        self._method = self['method'].string()

        # Collect number of unbinned, binned and On/Off observations in
        # observation container
        n_unbinned = 0
        n_binned   = 0
        n_onoff    = 0
        for obs in self.obs():
            if obs.classname() == 'GCTAObservation':
                if obs.eventtype() == 'CountsCube':
                    n_binned += 1
                else:
                    n_unbinned += 1
            elif obs.classname() == 'GCTAOnOffObservation':
                n_onoff += 1
        n_cta   = n_unbinned + n_binned + n_onoff
        n_other = self.obs().size() - n_cta

        # If spectrum method is not "NODES" then set spectrum method and
        # script mode according to type of observations
        if self._method != 'NODES':
            if n_other > 0:
                self._method = 'NODES'
            else:
                if n_unbinned == 0 and n_binned != 0 and n_onoff == 0:
                    self._binned_mode = True
                    self._method      = 'SLICE'
                elif n_unbinned == 0 and n_binned == 0 and n_onoff != 0:
                    self._onoff_mode = True
                    self._method      = 'SLICE'
                elif n_unbinned == 0 and n_binned != 0 and n_onoff != 0:
                    msg = 'Mix of binned and On/Off CTA observations found ' \
                          'in observation container. csscript does not support ' \
                          'this mix.'
                    raise RuntimeError(msg)
                elif n_unbinned != 0 and (n_binned != 0 or n_onoff != 0):
                    msg = 'Mix of unbinned and binned or On/Off CTA observations ' \
                          'found in observation container. csscript does not ' \
                          'support this mix.'
                    raise RuntimeError(msg)
                elif n_unbinned != 0:
                    self._method = 'SLICE'

        # Set ebounds
        self._set_ebounds()

        # Query other parameeters
        self['edisp'].boolean()
        self['calc_ulim'].boolean()
        self['calc_ts'].boolean()
        self['fix_bkg'].boolean()
        self['fix_srcs'].boolean()

        # Read ahead output parameters
        if self._read_ahead():
            self['outfile'].filename()

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)

        # Set number of processes for multiprocessing
        self._nthreads = mputils.nthreads(self)

        # Write spectrum method header and parameters
        self._log_header1(gammalib.TERSE, 'Spectrum method')
        self._log_value(gammalib.TERSE, 'Unbinned CTA observations', n_unbinned)
        self._log_value(gammalib.TERSE, 'Binned CTA observations', n_binned)
        self._log_value(gammalib.TERSE, 'On/off CTA observations', n_onoff)
        self._log_value(gammalib.TERSE, 'Other observations', n_other)

        # If there are a mix of CTA and non-CTA observations and the method
        # is 'SLICE' then log a warning that non-CTA observations will be
        # ignored
        warning = False
        if n_cta > 0 and n_other > 0 and self._method == 'SLICE':
            warning = True

        # If there are only non-CTA observations and the method is 'SLICE'
        # then stop now
        elif n_other > 0:
            if self._method == 'SLICE':
                msg = 'Selected "SLICE" method but none of the observations ' \
                      'is a CTA observation. Please select "AUTO" or "NODES" ' \
                      'if no CTA observation is provided.'
                raise RuntimeError(msg)
            else:
                self._method = 'NODES'

        # Log selected spectrum method
        self._log_value(gammalib.TERSE, 'Selected spectrum method', self._method)

        # Signal warning
        if warning:
            self._log_string(gammalib.TERSE, ' WARNING: Only CTA observation '
                             'can be handled with the "SLICE" method, all '
                             'non-CTA observation will be ignored.')

        # Return
        return

    def _set_ebounds(self):
        """
        Set energy boundaries
        """
        # If we are in binned or On/Off mode then align the energy boundaries
        # with the cube of RMF spectrum
        if self._binned_mode or self._onoff_mode:

            # Initialise energy boundaries for spectrum
            self._ebounds = gammalib.GEbounds()

            # Create energy boundaries according to user parameters
            ebounds = self._create_ebounds()

            # Extract binned energy boundaries from first observation in
            # container. This assumes that all observations have the same
            # energy binning. But even if this is not the case, the script
            # should work reasonably well since for each observation the
            # relevant energy bins will be selected.
            if self._binned_mode:
                cube_ebounds = self.obs()[0].events().ebounds()
            else:
                cube_ebounds = self.obs()[0].rmf().emeasured()

            # Loop over user energy boundaries and collect all energy bins
            # that overlap
            for i in range(ebounds.size()):

                # Extract minimum and maximum energy of user energy bin,
                # including some rounding tolerance
                emin = ebounds.emin(i).TeV() * 0.999 # Rounding tolerance
                emax = ebounds.emax(i).TeV() * 1.001 # Rounding tolerance

                # Set number of overlapping energy bins
                nbins = 0

                # Search first cube bin that is comprised within user energy
                # bin
                emin_value = -1.0
                for k in range(cube_ebounds.size()):
                    if cube_ebounds.emin(k).TeV() >= emin and \
                       cube_ebounds.emax(k).TeV() <= emax:
                        emin_value = cube_ebounds.emin(k).TeV()
                        break
                if emin_value < 0.0:
                    continue

                # Search last cube bin that is comprised within user energy bin
                for k in range(cube_ebounds.size()):
                    if cube_ebounds.emin(k).TeV() >= emin and \
                       cube_ebounds.emax(k).TeV() <= emax:
                        emax_value = cube_ebounds.emax(k).TeV()
                        nbins += 1

                # Append energy bin if there are overlapping bins in the
                # counts cube
                if nbins > 0:
                    self._ebounds.append(gammalib.GEnergy(emin_value, 'TeV'),
                                         gammalib.GEnergy(emax_value, 'TeV'))

            # Raise an exception if there are no overlapping layers
            if (len(self._ebounds) == 0):
                msg = 'Energy range ['+str(cube_ebounds.emin())+ \
                      ', '+str(cube_ebounds.emax())+'] of counts '+ \
                      'cube does not overlap with specified energy '+ \
                      'range ['+ \
                      str(ebounds.emin())+', '+str(ebounds.emax())+'].'+ \
                      ' Specify overlapping energy range.'
                raise RuntimeError(msg)

        # Unbinned mode
        else:
            self._ebounds = self._create_ebounds()

        # Return
        return

    def _log_spectral_binning(self):
        """
        Log spectral binning
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Spectral binning')

        # Log counts cube energy range for binned mode
        if self._binned_mode:
            cube_ebounds = self.obs()[0].events().ebounds()
            value = '%s - %s' % (str(cube_ebounds.emin()),
                                 str(cube_ebounds.emax()))
            self._log_value(gammalib.TERSE, 'Counts cube energy range', value)

        # Log RMF energy range for On/Off mode
        elif self._onoff_mode:
            etrue = self.obs()[0].rmf().etrue()
            ereco = self.obs()[0].rmf().emeasured()
            vtrue = '%s - %s' % (str(etrue.emin()), str(etrue.emax()))
            vreco = '%s - %s' % (str(ereco.emin()), str(ereco.emax()))
            self._log_value(gammalib.TERSE, 'RMF true energy range', vtrue)
            self._log_value(gammalib.TERSE, 'RMF measured energy range', vreco)

        # Log energy bins
        for i in range(self._ebounds.size()):
            value = '%s - %s' % (str(self._ebounds.emin(i)),
                                 str(self._ebounds.emax(i)))
            self._log_value(gammalib.TERSE, 'Bin %d' % (i+1), value)

        # Return
        return

    def _adjust_model_pars(self):
        """
        Adjust model parameters
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Adjust model parameters')

        # Adjust model parameters dependent on input user parameters
        for model in self.obs().models():

            # Initialise TS flag for all models to false
            model.tscalc(False)

            # Log model name
            self._log_header3(gammalib.EXPLICIT, model.name())

            # Deal with the source of interest    
            if model.name() == self['srcname'].string():

                # Fix all model parameters
                for par in model:
                    if par.is_free():
                        self._log_string(gammalib.EXPLICIT,
                                         ' Fixing "'+par.name()+'"')
                    par.fix()

                # Free the normalisation parameter which is assumed to be
                # the first spectral parameter
                normpar = model.spectral()[0]
                if normpar.is_fixed():
                    self._log_string(gammalib.EXPLICIT,
                                     ' Freeing "'+normpar.name()+'"')
                normpar.free()

                # Optionally compute Test Statistic value
                if self['calc_ts'].boolean():
                    model.tscalc(True)

            # Deal with background models
            elif self['fix_bkg'].boolean() and \
                 not model.classname() == 'GModelSky':
                for par in model:
                    if par.is_free():
                        self._log_string(gammalib.EXPLICIT,
                                         ' Fixing "'+par.name()+'"')
                    par.fix()

            # Deal with source models
            elif self['fix_srcs'].boolean() and \
                 model.classname() == 'GModelSky':
                for par in model:
                    if par.is_free():
                        self._log_string(gammalib.EXPLICIT,
                                         ' Fixing "'+par.name()+'"')
                    par.fix()

        # Return
        return

    def _set_replace_src_spectrum_by_nodes(self):
        """
        Replace source spectrum by node function
        """
        # Initialise model container
        models = gammalib.GModels()

        # Loop over model containers
        for model in self.obs().models():

            # If we deal with source model then replace the spectral model
            # by a node function
            if model.name() == self['srcname'].string():

                # Setup energies at log mean energies of bins
                energies = gammalib.GEnergies()
                for i in range(self._ebounds.size()):
                    energies.append(self._ebounds.elogmean(i))

                # Setup spectral node function
                spectrum = gammalib.GModelSpectralNodes(model.spectral(), energies)
                spectrum.autoscale()

                # Make sure that all nodes are positive. Autoscale all
                # parameters so that their nominal value is unity.
                for i in range(spectrum.nodes()):
                    par     = spectrum[i*2+1]
                    par.autoscale()
                    value   = par.value()
                    minimum = 1.0e-20 * value
                    if minimum <= 0.0:
                        minimum = 1.0e-20
                        if minimum < value:
                            value = minimum
                    par.value(value)
                    par.min(minimum)

                # Set spectral component of source model
                model.spectral(spectrum)

                # Make sure that TS computation is disabled (makes computation
                # faster)
                model.tscalc(False)

                # Append model
                models.append(model)

            # ... otherwise just append model
            else:
                models.append(model)

        # Put new model in observation containers
        self.obs().models(models)

        # Return
        return

    def _select_onoff_obs(self, obs, emin, emax):
        """
        Select an energy interval from one CTA On/Off observation

        Parameters
        ----------
        obs : `~gammalib.GCTAOnOffObservation`
            Minimum energy
        emin : `~gammalib.GEnergy()`
            Minimum energy
        emax : `~gammalib.GEnergy()`
            Maximum energy

        Returns
        -------
        obs : `~gammalib.GCTAOnOffObservation`
            CTA On/Off observation
        """
        # Select energy bins in etrue and ereco. All etrue energy bins are
        # selected. A 0.1% margin is added for reconstructed energies to
        # accomodate for rounding errors.
        etrue     = obs.rmf().etrue()
        ereco     = gammalib.GEbounds()
        itrue     = [i for i in range(obs.rmf().etrue().size())]
        ireco     = []
        for i in range(obs.rmf().emeasured().size()):
            ereco_bin_min = obs.rmf().emeasured().emin(i)
            ereco_bin_max = obs.rmf().emeasured().emax(i)
            if ereco_bin_min * 1.001 >= emin and ereco_bin_max * 0.999 <= emax:
                ereco.append(ereco_bin_min, ereco_bin_max)
                ireco.append(i)

        # Extract PHA
        pha_on  = gammalib.GPha(ereco)
        pha_off = gammalib.GPha(ereco)
        pha_on.exposure(obs.on_spec().exposure())
        pha_off.exposure(obs.on_spec().exposure())
        for idst, isrc in enumerate(ireco):
            # On
            pha_on[idst] = obs.on_spec()[isrc]
            pha_on.areascal(idst, obs.on_spec().areascal(isrc))
            pha_on.backscal(idst, obs.on_spec().backscal(isrc))
            # Off
            pha_off[idst] = obs.off_spec()[isrc]
            pha_off.areascal(idst, obs.off_spec().areascal(isrc))
            pha_off.backscal(idst, obs.off_spec().backscal(isrc))

        # Extract BACKRESP
        pha_backresp = obs.off_spec()['BACKRESP']
        backresp     = []
        for idst, isrc in enumerate(ireco):
            backresp.append(pha_backresp[isrc])
        pha_off.append('BACKRESP', backresp)

        # Extract ARF
        arf = gammalib.GArf(etrue)
        for idst, isrc in enumerate(itrue):
            arf[idst] = obs.arf()[isrc]

        # Extract RMF
        rmf = gammalib.GRmf(etrue, ereco)
        for idst_true, isrc_true in enumerate(itrue):
            for idst_reco, isrc_reco in enumerate(ireco):
                rmf[idst_true, idst_reco] = obs.rmf()[isrc_true, isrc_reco]

        # Set On/Off observations
        obsid      = obs.id()
        statistic  = obs.statistic()
        instrument = obs.instrument()
        obs = gammalib.GCTAOnOffObservation(pha_on, pha_off, arf, rmf)
        obs.id(obsid)
        obs.statistic(statistic)
        obs.instrument(instrument)

        # Return observation
        return obs

    def _select_obs(self, emin, emax):
        """
        Select observations for energy interval

        Parameters
        ----------
        emin : `~gammalib.GEnergy()`
            Minimum energy
        emax : `~gammalib.GEnergy()`
            Maximum energy

        Returns
        -------
        obs : `~gammalib.GObservations`
            Observation container
        """
        # Use ctcubemask for binned analysis
        if self._binned_mode:

            # Write header
            self._log_header3(gammalib.EXPLICIT, 'Filtering cube')

            # Select layers
            cubemask            = ctools.ctcubemask(self.obs())
            cubemask['regfile'] = 'NONE'
            cubemask['ra']      = 'UNDEFINED'
            cubemask['dec']     = 'UNDEFINED'
            cubemask['rad']     = 'UNDEFINED'
            cubemask['emin']    = emin.TeV()
            cubemask['emax']    = emax.TeV()

            # If chatter level is verbose and debugging is requested then
            # switch also on the debug model in ctcubemask
            if self._logVerbose() and self._logDebug():
                cubemask['debug'] = True

            # Select layers
            cubemask.run() 

            # Set new binned observation
            obs = cubemask.obs().copy()

        # Use ...
        elif self._onoff_mode:

            # Write header
            self._log_header3(gammalib.EXPLICIT, 'Filtering PHA, ARF and RMF')

            # Initialise observation container
            obs = gammalib.GObservations()

            # Loop over all input observations and select energy bins for
            # all On/Off observations
            for run in self.obs():
                if run.classname() == 'GCTAOnOffObservation':
                    obs.append(self._select_onoff_obs(run, emin, emax))

            # Append models
            obs.models(self.obs().models())

        # Use ctselect for unbinned analysis
        else:

            # Write header
            self._log_header3(gammalib.EXPLICIT, 'Selecting events')

            # Select events
            select = ctools.ctselect(self.obs())
            select['ra']   = 'UNDEFINED'
            select['dec']  = 'UNDEFINED'
            select['rad']  = 'UNDEFINED'
            select['emin'] = emin.TeV()
            select['emax'] = emax.TeV()
            select['tmin'] = 'UNDEFINED'
            select['tmax'] = 'UNDEFINED'

            # If chatter level is verbose and debugging is requested then
            # switch also on the debug model in ctselect
            if self._logVerbose() and self._logDebug():
                select['debug'] = True

            # Run ctselect
            select.run()

            # Retrieve observation
            obs = select.obs().copy()

        # Return observation container
        return obs

    def _fit_energy_bin(self, i):
        """
        Fit data for one energy bin

        Parameters
        ----------
        i : int
            Energy bin index

        Returns
        -------
        result : dict
            Dictionary with fit results
        """

        # Write header for energy bin
        self._log_header2(gammalib.EXPLICIT, 'Energy bin ' + str(i + 1))

        # Get energy boundaries
        emin      = self._ebounds.emin(i)
        emax      = self._ebounds.emax(i)
        elogmean  = self._ebounds.elogmean(i)

        # Select observations for energy bin
        obs = self._select_obs(emin, emax)

        # Initialise dictionary
        result = {'energy':      elogmean.TeV(),
                  'energy_low':  (elogmean - emin).TeV(),
                  'energy_high': (emax - elogmean).TeV(),
                  'flux':        0.0,
                  'flux_err':    0.0,
                  'TS':          0.0,
                  'ulimit':      0.0,
                  'Npred':       0.0}

        # Write header for fitting
        self._log_header3(gammalib.EXPLICIT, 'Performing fit in energy bin')

        # Setup maximum likelihood fit
        like = ctools.ctlike(obs)
        like['edisp']    = self['edisp'].boolean()
        like['nthreads'] = 1  # Avoids OpenMP conflict

        # If chatter level is verbose and debugging is requested then
        # switch also on the debug model in ctlike
        if self._logVerbose() and self._logDebug():
            like['debug'] = True

        # Perform maximum likelihood fit
        like.run()

        # Write model results for explicit chatter level
        self._log_string(gammalib.EXPLICIT, str(like.obs().models()))

        # Continue only if log-likelihood is non-zero
        if like.obs().logL() != 0.0:

            # Get results
            fitted_models = like.obs().models()
            source        = fitted_models[self['srcname'].string()]

            # Extract Test Statistic value
            if self['calc_ts'].boolean():
                result['TS'] = source.ts()

            # Compute Npred value (only works for unbinned analysis)
            if not self._binned_mode and not self._onoff_mode:
                for observation in like.obs():
                    result['Npred'] += observation.npred(source)

            # Compute upper flux limit
            ulimit_value = -1.0
            if self['calc_ulim'].boolean():

                # Logging information
                self._log_header3(gammalib.EXPLICIT,
                                  'Computing upper limit for energy bin')

                # Create upper limit object  
                ulimit = ctools.ctulimit(like.obs())
                ulimit['srcname'] = self['srcname'].string()
                ulimit['eref']    = elogmean.TeV()

                # If chatter level is verbose and debugging is requested
                # then switch also on the debug model in ctulimit
                if self._logVerbose() and self._logDebug():
                    ulimit['debug'] = True

                # Try to run upper limit and catch exceptions
                try:
                    ulimit.run()
                    ulimit_value = ulimit.diff_ulimit()
                except:
                    self._log_string(gammalib.EXPLICIT, 'Upper limit '
                                     'calculation failed.')
                    ulimit_value = -1.0

                # Compute upper limit
                if ulimit_value > 0.0:
                    result['ulimit'] = ulimit_value * elogmean.MeV() * \
                                       elogmean.MeV() * gammalib.MeV2erg

            # Compute differential flux and flux error
            fitted_flux = source.spectral().eval(elogmean)
            parvalue    = source.spectral()[0].value()
            if parvalue != 0.0:
                rel_error = source.spectral()[0].error() / parvalue
                e_flux    = fitted_flux * rel_error
            else:
                e_flux = 0.0

            # If the source model is a cube then multiply-in the cube
            # spectrum
            if source.spatial().classname() == 'GModelSpatialDiffuseCube':
                dir          = gammalib.GSkyDir()
                source.spatial().set_mc_cone(dir, 180.0)
                norm         = source.spatial().spectrum().eval(elogmean)
                fitted_flux *= norm
                e_flux      *= norm

            # Convert differential flux and flux error to nuFnu
            elogmean2          = elogmean.MeV() * elogmean.MeV()
            result['flux']     = fitted_flux * elogmean2 * gammalib.MeV2erg
            result['flux_err'] = e_flux      * elogmean2 * gammalib.MeV2erg

            # Log information
            value = '%e +/- %e' % (result['flux'], result['flux_err'])
            if self['calc_ulim'].boolean() and result['ulimit'] > 0.0:
                value += ' [< %e]' % (result['ulimit'])
            value += ' erg/cm2/s'
            if self['calc_ts'].boolean() and result['TS'] > 0.0:
                value += ' (TS = %.3f)' % (result['TS'])
            self._log_value(gammalib.TERSE, 'Bin '+str(i+1), value)

        # ... otherwise if logL is zero then signal that bin is
        # skipped
        else:
            value = 'Likelihood is zero. Bin is skipped.'
            self._log_value(gammalib.TERSE, 'Bin '+str(i+1), value)

        # Return result
        return result

    def _fit_energy_bins(self):
        """
        Fit model to energy bins

        Returns
        -------
        results : list of dict
            List of dictionaries with fit results
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Generate spectrum')
        self._log_string(gammalib.TERSE, str(self._ebounds))

        # Initialise results
        results = []

        # If more than a single thread is requested then use multiprocessing
        if self._nthreads > 1:

            # Compute energy bins
            args        = [(self, '_fit_energy_bin', i)
                           for i in range(self._ebounds.size())]
            poolresults = mputils.process(self._nthreads, mputils.mpfunc, args)

            # Construct results
            for i in range(self._ebounds.size()):
                results.append(poolresults[i][0])
                self._log_string(gammalib.TERSE, poolresults[i][1]['log'], False)

        # Otherwise, loop over energy bins
        else:
            for i in range(self._ebounds.size()):

                # Fit energy bin
                result = self._fit_energy_bin(i)

                # Append results
                results.append(result)

        # Return results
        return results

    def _fit_model(self):
        """
        Fit model to observations

        Returns
        -------
        results : list of dict
            List of dictionaries with fit results
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Generate spectrum')

        # Write header for fitting
        self._log_header3(gammalib.EXPLICIT, 'Performing model fit')

        # Perform maximum likelihood fit
        like          = ctools.ctlike(self.obs())
        like['edisp'] = self['edisp'].boolean()
        like.run()

        # Initialise fit results
        results = []

        # Extract fit results
        model    = like.obs().models()[self['srcname'].string()]
        spectrum = model.spectral()
        logL0    = like.obs().logL()

        # Write model results for explicit chatter level
        self._log_string(gammalib.EXPLICIT, str(like.obs().models()))

        # Loop over all nodes
        for i in range(spectrum.nodes()):

            # Get energy boundaries
            emin      = self._ebounds.emin(i)
            emax      = self._ebounds.emax(i)
            elogmean  = self._ebounds.elogmean(i)

            # Initialise dictionary
            result = {'energy':      elogmean.TeV(),
                      'energy_low':  (elogmean - emin).TeV(),
                      'energy_high': (emax - elogmean).TeV(),
                      'flux':        0.0,
                      'flux_err':    0.0,
                      'TS':          0.0,
                      'ulimit':      0.0,
                      'Npred':       0.0}

            # Convert differential flux and flux error to nuFnu
            norm               = elogmean.MeV() * elogmean.MeV()  * gammalib.MeV2erg
            result['flux']     = spectrum[i*2+1].value() * norm
            result['flux_err'] = spectrum[i*2+1].error() * norm

            # Compute upper flux limit
            ulimit_value = -1.0
            if self['calc_ulim'].boolean():

                # Logging information
                self._log_header3(gammalib.EXPLICIT,
                                  'Computing upper limit for node energy %f TeV' %
                                  result['energy'])

                # Copy observation container
                obs = like.obs().copy()

                # Fix intensities of all nodes
                spectral = obs.models()[self['srcname'].string()].spectral()
                for par in spectral:
                    par.fix()

                # Create upper limit object  
                ulimit = ctools.ctulimit(obs)
                ulimit['srcname'] = self['srcname'].string()
                ulimit['parname'] = 'Intensity%d' % i
                ulimit['eref']    = elogmean.TeV()
                ulimit['tol']     = 1.0e-3

                # Try to run upper limit and catch exceptions
                try:
                    ulimit.run()
                    ulimit_value = ulimit.diff_ulimit()
                except:
                    self._log_string(gammalib.EXPLICIT, 'Upper limit '
                                     'calculation failed.')
                    ulimit_value = -1.0

                # Compute upper limit
                if ulimit_value > 0.0:
                    result['ulimit'] = ulimit_value * elogmean.MeV() * \
                                       elogmean.MeV() * gammalib.MeV2erg

            # Compute TS
            if self['calc_ts'].boolean():

                # Copy observation container
                obs = like.obs().copy()

                # Set intensity of node to tiny value by scaling the value
                # by a factor 1e-8.
                par = obs.models()[self['srcname'].string()].spectral()[i*2+1]
                par.autoscale()
                par.factor_min(1.0e-8)
                par.factor_value(1.0e-8)
                par.autoscale()
                par.fix()

                # Perform maximum likelihood fit
                tslike          = ctools.ctlike(obs)
                tslike['edisp'] = self['edisp'].boolean()
                tslike.run()

                # Store Test Statistic
                model        = tslike.obs().models()[self['srcname'].string()]
                logL1        = tslike.obs().logL()
                result['TS'] = 2.0 * (logL1 - logL0)

            # Log information
            value = '%e +/- %e' % (result['flux'], result['flux_err'])
            if self['calc_ulim'].boolean() and result['ulimit'] > 0.0:
                value += ' [< %e]' % (result['ulimit'])
            value += ' erg/cm2/s'
            if self['calc_ts'].boolean() and result['TS'] > 0.0:
                value += ' (TS = %.3f)' % (result['TS'])
            self._log_value(gammalib.TERSE, 'Bin '+str(i+1), value)

            # Append results
            results.append(result)

        # Return results
        return results

    def _create_fits(self, results):
        """
        Create FITS file

        Parameters
        ----------
        results : list of dict
            List of result dictionaries
        """
        # Create FITS table columns
        nrows        = self._ebounds.size()
        energy       = gammalib.GFitsTableDoubleCol('Energy', nrows)
        energy_low   = gammalib.GFitsTableDoubleCol('ed_Energy', nrows)
        energy_high  = gammalib.GFitsTableDoubleCol('eu_Energy', nrows)
        flux         = gammalib.GFitsTableDoubleCol('Flux', nrows)
        flux_err     = gammalib.GFitsTableDoubleCol('e_Flux', nrows)
        TSvalues     = gammalib.GFitsTableDoubleCol('TS', nrows)
        ulim_values  = gammalib.GFitsTableDoubleCol('UpperLimit', nrows)
        Npred_values = gammalib.GFitsTableDoubleCol('Npred', nrows)
        energy.unit('TeV')
        energy_low.unit('TeV')
        energy_high.unit('TeV')
        flux.unit('erg/cm2/s')
        flux_err.unit('erg/cm2/s')
        ulim_values.unit('erg/cm2/s')

        # File FITS table columns
        for i, result in enumerate(results):
            energy[i]       = result['energy']
            energy_low[i]   = result['energy_low']
            energy_high[i]  = result['energy_high']
            flux[i]         = result['flux']
            flux_err[i]     = result['flux_err']
            TSvalues[i]     = result['TS']
            ulim_values[i]  = result['ulimit']
            Npred_values[i] = result['Npred']

        # Initialise FITS Table with extension "SPECTRUM"
        table = gammalib.GFitsBinTable(nrows)
        table.extname('SPECTRUM')

        # Add Header for compatibility with gammalib.GMWLSpectrum
        table.card('INSTRUME', 'CTA', 'Name of Instrument')
        table.card('TELESCOP', 'CTA', 'Name of Telescope')

        # Append filled columns to fits table    
        table.append(energy)
        table.append(energy_low)
        table.append(energy_high)
        table.append(flux)
        table.append(flux_err)
        table.append(TSvalues)
        table.append(ulim_values)
        table.append(Npred_values)

        # Create the FITS file now
        self._fits = gammalib.GFits()
        self._fits.append(table)

        # Return
        return


    # Public methods
    def run(self):
        """
        Run the script
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()

        # Write input observation container into logger
        self._log_observations(gammalib.NORMAL, self.obs(), 'Input observation')

        # Write spectral binning into logger
        self._log_spectral_binning()

        # Adjust model parameters dependent on input user parameters
        self._adjust_model_pars()

        # Case A: "SLICE" method
        if self._method == 'SLICE':

            # Fit energy bins
            results = self._fit_energy_bins()

        # Case B: "NODES" method
        elif self._method == 'NODES':

            # Replace source spectrum by nodes function
            self._set_replace_src_spectrum_by_nodes()

            # Fit model
            results = self._fit_model()

        # Create FITS file
        self._create_fits(results)

        # Optionally publish spectrum
        if self['publish'].boolean():
            self.publish()

        # Return
        return

    def save(self):
        """
        Save spectrum
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Save spectrum')

        # Continue only if FITS file is valid
        if self._fits != None:

            # Get outmap parameter
            outfile = self['outfile'].filename()

            # Log file name
            self._log_value(gammalib.NORMAL, 'Spectrum file', outfile.url())

            # Save spectrum
            self._fits.saveto(outfile, self['clobber'].boolean())

        # Return
        return

    def publish(self, name=''):
        """
        Publish spectrum

        Parameters
        ----------
        name : str, optional
            Name of spectrum
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Publish spectrum')

        # Continue only if FITS file is valid
        if self._fits != None:

            # Continue only if spectrum is valid
            if self._fits.contains('SPECTRUM'):

                # Set default name is user name is empty
                if not name:
                    user_name = self._name()
                else:
                    user_name = name

                # Log file name
                self._log_value(gammalib.NORMAL, 'Spectrum name', user_name)

                # Publish spectrum
                self._fits.publish('SPECTRUM', user_name)

        # Return
        return

    def spectrum(self):
        """
        Return spectrum FITS file

        Returns:
            FITS file containing spectrum
        """
        # Return
        return self._fits

    def models(self, models):
        """
        Set model
        """
        # Copy models
        self.obs().models(models.clone())

        # Return
        return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = csspec(sys.argv)

    # Execute application
    app.execute()



#! /usr/bin/env python
# ==========================================================================
# Display spectrum generated by csspec
#
# Copyright (C) 2015-2019 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
try:
    import matplotlib.pyplot as plt
    plt.figure()
    plt.close()
except (ImportError, RuntimeError):
    print('This script needs the "matplotlib" module')
    sys.exit()
import gammalib
import cscripts


# ============= #
# Plot spectrum #
# ============= #
def plot_spectrum(filename, plotfile):
    """
    Plot spectrum

    Parameters
    ----------
    filename : str
        Name of spectrum FITS file
    plotfile : str
        Plot file name
    """
    # Read spectrum file    
    fits     = gammalib.GFits(filename)
    table    = fits.table(1)
    c_energy = table['Energy']
    c_ed     = table['ed_Energy']
    c_eu     = table['eu_Energy']
    c_flux   = table['Flux']
    c_eflux  = table['e_Flux']
    c_ts     = table['TS']
    c_upper  = table['UpperLimit']

    # Initialise arrays to be filled
    energies    = []
    flux        = []
    ed_engs     = []
    eu_engs     = []
    e_flux      = []
    ul_energies = []
    ul_ed_engs  = []
    ul_eu_engs  = []
    ul_flux     = []

    # Loop over rows of the file
    nrows = table.nrows()
    for row in range(nrows):

        # Get Test Statistic, flux and flux error
        ts    = c_ts.real(row)
        flx   = c_flux.real(row)
        e_flx = c_eflux.real(row)

        # If Test Statistic is larger than 9 and flux error is smaller than
        # flux then append flux plots ...
        if ts > 9.0 and e_flx < flx:
            energies.append(c_energy.real(row))
            flux.append(c_flux.real(row))
            ed_engs.append(c_ed.real(row))
            eu_engs.append(c_eu.real(row))
            e_flux.append(c_eflux.real(row))

        # ... otherwise append upper limit
        else:
            ul_energies.append(c_energy.real(row))
            ul_flux.append(c_upper.real(row))
            ul_ed_engs.append(c_ed.real(row))
            ul_eu_engs.append(c_eu.real(row))

    # Set upper limit errors
    yerr = [0.6 * x for x in ul_flux]

    # Plot the spectrum 
    plt.figure()
    plt.loglog()
    plt.grid()
    plt.errorbar(energies, flux, yerr=e_flux, xerr=[ed_engs, eu_engs],
                 fmt='ro')
    plt.errorbar(ul_energies, ul_flux, xerr=[ul_ed_engs, ul_eu_engs],
                 yerr=yerr, uplims=True, fmt='ro')
    plt.xlabel('Energy (TeV)')
    plt.ylabel(r'E$^2$ $\times$ dN/dE (erg cm$^{-2}$ s$^{-1}$)')

    # Show figure
    if len(plotfile) > 0:
        plt.savefig(plotfile)
    else:
        plt.show()

    # Return
    return


# ============= #
# Show spectrum #
# ============= #
def show_spectrum():
    """
    Show spectrum
    """
    # Set usage string
    usage = 'show_spectrum.py [-p plotfile] [file]'

    # Set default options
    options = [{'option': '-p', 'value': ''}]

    # Get arguments and options from command line arguments
    args, options = cscripts.ioutils.get_args_options(options, usage)

    # Extract script parameters from options
    plotfile = options[0]['value']

    # Show spectrum
    plot_spectrum(args[0], plotfile)

    # Return
    return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Show spectrum
    show_spectrum()



#! /usr/bin/env python
# ==========================================================================
# Show a spectrum in PHA format
#
# Copyright (C) 2013-2016 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import math
import gammalib
import cscripts
try:
    import matplotlib.pyplot as plt
    plt.figure()
    plt.close()
except (ImportError, RuntimeError):
    print('This script needs the "matplotlib" module')
    sys.exit()


# ================= #
# Plot PHA spectrum #
# ================= #
def plot_pha(pha, plotfile):
    """
    Plot PHA spectrum

    Parameters
    ----------
    spectrum : `~gammalib.GPha`
        PHA file
    plotfile : str
        Plot file name
    """
    # Create figure
    plt.figure(1)
    plt.title('PHA spectrum (' + pha.filename() + ')')

    # Generate energy vector
    energy   = []
    channels = False
    if pha.ebounds().size() > 0:
        for i in range(pha.ebounds().size()):
            energy.append(pha.ebounds().elogmean(i).TeV())
    else:
        for i in range(pha.size()):
            energy.append(float(i+1))
            channels = True

    # Generate data and error vector
    counts = [0.0 for i in range(pha.size())]
    for i in range(pha.size()):
        counts[i] = pha[i]
    error = [math.sqrt(c) for c in counts]

    # Plot data
    if channels:
        plt.semilogy(energy, counts, 'ro')
    else:
        plt.loglog(energy, counts, 'ro')
    plt.errorbar(energy, counts, error, fmt='ro', ecolor='r')

    # Set axes
    if channels:
        plt.xlabel('Channels')
    else:
        plt.xlabel('Energy (TeV)')
    plt.ylabel('Counts')

    # Show spectrum or save it into file
    if len(plotfile) > 0:
        plt.savefig(plotfile)
    else:
        plt.show()

    # Return
    return


# ================= #
# Show PHA spectrum #
# ================= #
def show_pha():
    """
    Show PHA spectrum
    """
    # Set usage string
    usage = 'show_pha.py [-p plotfile] file'

    # Set default options
    options = [{'option': '-p', 'value': ''}]

    # Get arguments and options from command line arguments
    args, options = cscripts.ioutils.get_args_options(options, usage)

    # Extract script parameters from options
    plotfile = options[0]['value']

    # Load PHA spectrum
    pha = gammalib.GPha(args[0])

    # Plot PHA spectrum
    plot_pha(pha, plotfile)

    # Return
    return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Show PHA spectrum
    show_pha()



# ==========================================================================
# Utility functions for observation handling
#
# Copyright (C) 2011-2019 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import math
import gammalib
import ctools
import cscripts


# ===================== #
# Simulate observations #
# ===================== #
def sim(obs, log=False, debug=False, chatter=2, edisp=False, seed=0,
        emin=None, emax=None, nbins=0, onsrc=None, onrad=0.2, addbounds=False,
        binsz=0.05, npix=200, proj='TAN', coord='GAL', nthreads=0):
    """
    Simulate events for all observations in the container

    Simulate events for all observations using ctobssim. If the number of
    energy bins is positive, the events are binned into a counts cube using
    ctbin. If multiple observations are simulated, the counts cube is a
    stacked cube and the corresponding response cubes are computed using
    ctexpcube, ctpsfcube, ctbkgcube and optionally ctedispcube. The response
    cubes are attached to the first observation in the container, which
    normally is the observation with the counts cube.

    Parameters
    ----------
    obs : `~gammalib.GObservations`
        Observation container without events
    log : bool, optional
        Create log file(s)
    debug : bool, optional
        Create console dump?
    chatter : int, optional
        Chatter level
    edisp : bool, optional
        Apply energy dispersion?
    seed : int, optional
        Seed value for simulations
    emin : float, optional
        Minimum energy of counts cube for binned (TeV)
    emax : float, optional
        Maximum energy of counts cube for binned (TeV)
    nbins : int, optional
        Number of energy bins (0=unbinned)
    onsrc : str, optional
        Name of source for On region (None if no On/Off obs. is used)
    onrad : float, optional
        Radius for On region (deg)
    addbounds : bool, optional
        Add boundaries at observation energies
    binsz : float, optional
        Pixel size for binned simulation (deg/pixel)
    npix : int, optional
        Number of pixels in X and Y for binned simulation
    proj : str, optional
        Projection for binned simulation
    coord : str, optional
        Coordinate system for binned simulation
    nthreads : str, optional
        Number of parallel processes for On/Off spectra computation (0=all available CPUs)

    Returns
    -------
    obs : `~gammalib.GObservations`
        Observation container filled with simulated events
    """
    # Allocate ctobssim application and set parameters
    obssim = ctools.ctobssim(obs)
    obssim['seed']     = seed
    obssim['edisp']    = edisp
    obssim['nthreads'] = nthreads
    obssim['chatter']  = chatter
    obssim['debug']    = debug

    # Optionally open the log file
    if log:
        obssim.logFileOpen()

    # Run ctobssim application. This will loop over all observations in the
    # container and simulation the events for each observation. Note that
    # events are not added together, they still apply to each observation
    # separately.
    obssim.run()

    # Binned option?
    if nbins > 0:

        # If energy boundaries are not given then determine the minimum and
        # the maximum energies from all observations and use these values
        # as energy boundaries. The energy boundaries are given in TeV.
        if emin == None or emax == None:
            emin = 1.0e30
            emax = 0.0
            for run in obssim.obs():
                emin = min(run.events().ebounds().emin().TeV(), emin)
                emax = max(run.events().ebounds().emax().TeV(), emax)

        # If a On source is specified then create On/Off observations
        if onsrc != None:

            # Extract source position from model
            model = obssim.obs().models()[onsrc]
            ra    = model['RA'].value()
            dec   = model['DEC'].value()

            # Allocate csphagen application and set parameters
            phagen = cscripts.csphagen(obssim.obs())
            phagen['inmodel']     = 'NONE'
            phagen['srcname']     = onsrc
            phagen['ebinalg']     = 'LOG'
            phagen['emin']        = emin
            phagen['emax']        = emax
            phagen['enumbins']    = nbins
            phagen['coordsys']    = 'CEL'
            phagen['ra']          = ra
            phagen['dec']         = dec
            phagen['rad']         = onrad
            phagen['stack']       = False
            phagen['inexclusion'] = 'NONE'
            phagen['bkgmethod']   = 'REFLECTED'
            phagen['nthreads']    = nthreads

            # Optionally open the log file
            if log:
                phagen.logFileOpen()

            # Run csphagen application
            phagen.run()

            # Make a deep copy of the observation that will be returned
            # (the csphagen object will go out of scope one the function is
            # left)
            obs = phagen.obs().copy()

        # ... otherwise use binned observations
        else:

            # Allocate ctbin application and set parameters
            binning = ctools.ctbin(obssim.obs())
            binning['ebinalg']  = 'LOG'
            binning['emin']     = emin
            binning['emax']     = emax
            binning['enumbins'] = nbins
            binning['usepnt']   = True # Use pointing for map centre
            binning['nxpix']    = npix
            binning['nypix']    = npix
            binning['binsz']    = binsz
            binning['coordsys'] = coord
            binning['proj']     = proj
            binning['nthreads'] = nthreads
            binning['chatter']  = chatter
            binning['debug']    = debug

            # Optionally open the log file
            if log:
                binning.logFileOpen()

            # Run ctbin application. This will loop over all observations in
            # the container and bin the events in counts maps
            binning.run()

            # If we have multiple input observations then create stacked response
            # cubes and append them to the observation
            if len(obssim.obs()) > 1:

                # Get counts cube. The counts cube is needed to obtained a
                # proper background cube.
                cntcube = binning.cube()

                # Get stacked response
                response = get_stacked_response(obssim.obs(), cntcube,
                                                edisp=edisp,
                                                addbounds=addbounds,
                                                log=log, debug=debug,
                                                chatter=chatter)

                # Set stacked response
                if edisp:
                    binning.obs()[0].response(response['expcube'],
                                              response['psfcube'],
                                              response['edispcube'],
                                              response['bkgcube'])
                else:
                    binning.obs()[0].response(response['expcube'],
                                              response['psfcube'],
                                              response['bkgcube'])

                # Set new models
                binning.obs().models(response['models'])

            # Make a deep copy of the observation that will be returned
            # (the ctbin object will go out of scope one the function is
            # left)
            obs = binning.obs().copy()

    else:

        # Make a deep copy of the observation that will be returned
        # (the ctobssim object will go out of scope one the function is
        # left)
        obs = obssim.obs().copy()

    # Delete the simulation
    del obssim

    # Return observation container
    return obs


# ======================= #
# Set one CTA observation #
# ======================= #
def set_obs(pntdir, tstart=0.0, duration=1800.0, deadc=0.98, \
            emin=0.1, emax=100.0, rad=5.0, \
            irf='South_50h', caldb='prod2', obsid='000000', mjdref=51544.5):
    """
    Set a single CTA observation

    The function sets a single CTA observation containing an empty CTA
    event list. By looping over this function CTA observations can be
    added to the observation container.

    Parameters
    ----------
    pntdir : `~gammalib.GSkyDir`
        Pointing direction
    tstart : float, optional
        Start time (s)
    duration : float, optional
        Duration of observation (s)
    deadc : float, optional
        Deadtime correction factor
    emin : float, optional
        Minimum event energy (TeV)
    emax : float, optional
        Maximum event energy (TeV)
    rad : float, optional
        ROI radius used for analysis (deg)
    irf : str, optional
        Instrument response function
    caldb : str, optional
        Calibration database path
    obsid : str, optional
        Observation identifier

    Returns
    -------
    obs : `~gammalib.GCTAObservation`
        CTA observation
    """
    # Allocate CTA observation
    obs = gammalib.GCTAObservation()

    # Set CTA calibration database
    db = gammalib.GCaldb()
    if (gammalib.dir_exists(caldb)):
        db.rootdir(caldb)
    else:
        db.open('cta', caldb)

    # Set pointing direction for CTA observation
    pnt = gammalib.GCTAPointing()
    pnt.dir(pntdir)
    obs.pointing(pnt)

    # Set ROI
    roi     = gammalib.GCTARoi()
    instdir = gammalib.GCTAInstDir()
    instdir.dir(pntdir)
    roi.centre(instdir)
    roi.radius(rad)

    # Set GTI
    ref = gammalib.GTimeReference(mjdref,'s','TT','LOCAL')
    gti = gammalib.GGti(ref)
    gti.append(gammalib.GTime(tstart, ref),
               gammalib.GTime(tstart+duration, ref))

    # Set energy boundaries
    ebounds = gammalib.GEbounds(gammalib.GEnergy(emin, 'TeV'),
                                gammalib.GEnergy(emax, 'TeV'))

    # Allocate event list
    events = gammalib.GCTAEventList()

    # Set ROI, GTI and energy boundaries for event list
    events.roi(roi)
    events.gti(gti)
    events.ebounds(ebounds)

    # Set the event list as the events for CTA observation
    obs.events(events)

    # Set instrument response for CTA observation
    obs.response(irf, db)

    # Set ontime, livetime, and deadtime correction factor for CTA observation
    obs.ontime(duration)
    obs.livetime(duration*deadc)
    obs.deadc(deadc)
    obs.id(obsid)

    # Return CTA observation
    return obs


# ============================ #
# Set list of CTA observations #
# ============================ #
def set_obs_list(obsdeflist, tstart=0.0, duration=1800.0, deadc=0.98, \
                 emin=0.1, emax=100.0, rad=5.0, \
                 irf='South_50h', caldb='prod2'):
    """
    Returns an observation container filled with a list of CTA observations

    The list is defined by the obsdeflist parameter which is a dictionnary
    containing the mandatory keywords 'ra' and 'dec' that specify the
    pointing direction for a given observation. Optional keyword give control
    over other observation proposerties, such as duration, deadtime correction,
    energy range, etc. If an optional keyword is not specified, the function
    keyword is used instead.

    Parameters
    ----------
    obsdeflist : list of dict
        Observation definition list
    tstart : float, optional
        Start time (s)
    duration : float, optional
        Duration of observation (s)
    deadc : float, optional
        Deadtime correction factor
    emin : float, optional
        Minimum event energy (TeV)
    emax : float, optional
        Maximum event energy (TeV)
    rad : float, optional
        ROI radius used for analysis (deg)
    irf : str, optional
        Instrument response function
    caldb : str, optional
        Calibration database path

    Returns
    -------
    obs : `~gammalib.GObservations`
        Observation container filled with CTA observation
    """
    # Initialise empty observation container
    obs = gammalib.GObservations()

    # Initialise first time and identifier
    obs_start = tstart
    obs_id    = 1

    # Loop over observation definition list
    for obsdef in obsdeflist:

        # Set pointing direction
        pntdir = gammalib.GSkyDir()
        pntdir.radec_deg(obsdef['ra'], obsdef['dec'])

        # Generate identifier string
        obsid = '%6.6d' % obs_id

        # Set one CTA observation
        obs_cta = set_obs(pntdir,
                          tstart=obs_start,
                          duration=obsdef.setdefault('duration', duration),
                          deadc=obsdef.setdefault('deadc', deadc),
                          emin=obsdef.setdefault('emin', emin),
                          emax=obsdef.setdefault('emax', emax),
                          rad=obsdef.setdefault('rad', rad),
                          irf=obsdef.setdefault('irf', irf),
                          caldb=obsdef.setdefault('caldb', caldb),
                          obsid=obsid)

        # Append to container
        obs.append(obs_cta)

        # Update start time and identifier
        obs_start += duration
        obs_id    += 1

    # Return observation container
    return obs


# ============================ #
# Set CTA observation patterns #
# ============================ #
def set_obs_patterns(pattern, ra=83.6331, dec=22.0145, offset=1.5):
    """
    Sets a number of standard patterns

    Parameters
    ----------
    pattern : str
        Observation pattern ("single", "four")
    ra : float, optional
        Right Ascension of pattern centre (deg)
    dec : float, optional
        Declination of pattern centre (deg)
    offset : float, optional
        Offset from pattern centre (deg)

    Returns
    -------
    obsdeflist : list
        Observation definition list
    """
    # Initialise observation definition list
    obsdeflist = []

    # If the pattern is a single observation then append the Right Ascension
    # and Declination to the observation definition list
    if pattern == 'single':
        obsdef = {'ra': ra, 'dec': dec}
        obsdeflist.append(obsdef)

    # ... otherwise, if the pattern is four observations then append four
    # observations offset by a certain amount from the pattern centre to the
    # observation definition list
    elif pattern == 'four':

        # Set pattern centre
        centre = gammalib.GSkyDir()
        centre.radec_deg(ra, dec)

        # Append pointings
        for phi in [0.0, 90.0, 180.0, 270.0]:
            pntdir = centre.copy()
            pntdir.rotate_deg(phi, offset)
            obsdeflist.append({'ra': pntdir.ra_deg(), 'dec': pntdir.dec_deg()})

    # ... otherwise raise an exception since we have an unknown pattern
    else:
        msg = 'Observation pattern "%s" not recognized.' % (pattern)
        raise RuntimeError(msg)

    # Return observation definition list
    return obsdeflist


# ====================================================== #
# Set observation container filled with CTA observations #
# ====================================================== #
def set_observations(ra, dec, rad, tstart, duration, emin, emax, irf, caldb,
                     deadc=0.98, pattern='single', offset=1.5):
    """
    Set an observation container filled with CTA observations

    Parameters
    ----------
    ra : float
        Right Ascension of pattern centre (deg)
    dec : float
        Declination of pattern centre (deg)
    rad : float
        ROI radius used for analysis (deg)
    tstart : float
        Start time of observation (s)
    duration : float
        Duration of each observation (s)
    emin : float, optional
        Minimum event energy (TeV)
    emax : float, optional
        Maximum event energy (TeV)
    irf : str
        Instrument response function
    caldb : str
        Calibration database path
    deadc : float, optional
        Deadtime correction factor
    pattern : str, optional
        Observation pattern ("single", "four")
    offset : float, optional
        Offset from pattern centre (deg)

    Returns
    -------
    obs : `~gammalib.GObservations()
        Observation container
    """
    # Setup observation definition list
    obsdeflist = set_obs_patterns(pattern, offset=offset, ra=ra, dec=dec)

    # Create list of observations
    obs = set_obs_list(obsdeflist, tstart=tstart, duration=duration,
                       emin=emin, emax=emax, rad=rad, irf=irf, caldb=caldb,
                       deadc=deadc)

    # Return observation container
    return obs


# ==================== #
# Get stacked response #
# ==================== #
def get_stacked_response(obs, cntcube, edisp=False, addbounds=False,
                         log=False, debug=False, chatter=2):
    """
    Get stacked response cubes

    The number of energies bins are set to at least 30 bins per decade, unless
    the counts cube has more energy bins per decade.

    Parameters
    ----------
    obs : `~gammalib.GObservations`
        Observation container
    cntcube : `~gammalib.GCTAEventCube`
        Counts cube
    edisp : bool, optional
        Apply energy dispersion?
    addbounds : bool, optional
        Add boundaries at observation energies
    log : bool, optional
        Create log file(s)
    debug : bool, optional
        Create console dump?
    chatter : int, optional
        Chatter level

    Returns
    -------
    result : dict
        Dictionary of response cubes
    """
    # Derive binning parameters from counts cube
    xref     = cntcube.counts().projection().crval(0)
    yref     = cntcube.counts().projection().crval(1)
    binsz    = cntcube.counts().projection().cdelt(1)
    coordsys = cntcube.counts().projection().coordsys()
    proj     = cntcube.counts().projection().code()
    nxpix    = cntcube.nx()
    nypix    = cntcube.ny()
    emin     = cntcube.emin().TeV()
    emax     = cntcube.emax().TeV()
    ebins    = cntcube.ebins()

    # Translate coordinate system
    if coordsys == 'EQU':
        coordsys = 'CEL'

    # Set energy limits, with larger etrue limits for the case that energy
    # dispersion is requested
    if edisp:
        emin *= 0.5
        emax *= 1.5

    # Set number of energy bins to at least 30 per energy decade
    enumbins = int((math.log10(emax) - math.log10(emin)) * 30.0)
    if ebins > enumbins:
        enumbins = ebins

    # Compute spatial binning for point spread function and energy dispersion
    # cubes. The spatial binning is 10 times coarser than the spatial binning
    # of the exposure and background cubes. At least 2 spatial are required.
    psf_binsz = 10.0 * binsz
    psf_nxpix = max(nxpix // 10, 2)  # Make sure result is int
    psf_nypix = max(nypix // 10, 2)  # Make sure result is int

    # Create exposure cube
    expcube = ctools.ctexpcube(obs)
    expcube['incube']    = 'NONE'
    expcube['ebinalg']   = 'LOG'
    expcube['xref']      = xref
    expcube['yref']      = yref
    expcube['binsz']     = binsz
    expcube['nxpix']     = nxpix
    expcube['nypix']     = nypix
    expcube['enumbins']  = enumbins
    expcube['emin']      = emin
    expcube['emax']      = emax
    expcube['coordsys']  = coordsys
    expcube['proj']      = proj
    expcube['addbounds'] = addbounds
    expcube['debug']     = debug
    expcube['chatter']   = chatter
    if log:
        expcube.logFileOpen()
    expcube.run()

    # Create point spread function cube
    psfcube = ctools.ctpsfcube(obs)
    psfcube['incube']    = 'NONE'
    psfcube['ebinalg']   = 'LOG'
    psfcube['xref']      = xref
    psfcube['yref']      = yref
    psfcube['binsz']     = psf_binsz
    psfcube['nxpix']     = psf_nxpix
    psfcube['nypix']     = psf_nypix
    psfcube['enumbins']  = enumbins
    psfcube['emin']      = emin
    psfcube['emax']      = emax
    psfcube['coordsys']  = coordsys
    psfcube['proj']      = proj
    psfcube['addbounds'] = addbounds
    psfcube['debug']     = debug
    psfcube['chatter']   = chatter
    if log:
        psfcube.logFileOpen()
    psfcube.run()

    # Create background cube. Note that we use the same energy binning as
    # for the counts cube since no interpolation should be actually done
    # for the background cube.
    bkgcube = ctools.ctbkgcube(obs)
    bkgcube.cntcube(cntcube)
    bkgcube['debug']   = debug
    bkgcube['chatter'] = chatter
    if log:
        bkgcube.logFileOpen()
    bkgcube.run()

    # If energy dispersion is requested then create energy dispersion cube
    if edisp:
        edispcube = ctools.ctedispcube(obs)
        edispcube['incube']    = 'NONE'
        edispcube['ebinalg']   = 'LOG'
        edispcube['xref']      = xref
        edispcube['yref']      = yref
        edispcube['binsz']     = psf_binsz
        edispcube['nxpix']     = psf_nxpix
        edispcube['nypix']     = psf_nypix
        edispcube['enumbins']  = enumbins
        edispcube['emin']      = emin
        edispcube['emax']      = emax
        edispcube['coordsys']  = coordsys
        edispcube['proj']      = proj
        edispcube['addbounds'] = addbounds
        edispcube['debug']     = debug
        edispcube['chatter']   = chatter
        if log:
            edispcube.logFileOpen()
        edispcube.run()

    # Build response dictionary
    response = {}
    response['expcube'] = expcube.expcube().copy()
    response['psfcube'] = psfcube.psfcube().copy()
    response['bkgcube'] = bkgcube.bkgcube().copy()
    response['models']  = bkgcube.models().copy()
    if edisp:
        response['edispcube'] = edispcube.edispcube().copy()

    # Return response cubes
    return response


# ================================= #
# Get stacked observation container #
# ================================= #
def get_stacked_obs(cls, obs, nthreads=0):
    """
    Bin an observation and return an observation container with a single
    binned observation

    Parameters
    ----------
    cls : `~ctools.cscript`
        cscript class
    obs : `~gammalib.GObservations`
        Observation container
    nthreads : str, optional
        Number of parallel processes for On/Off spectra computation (0=all available CPUs)

    Returns
    -------
    obs : `~gammalib.GObservations`
        Observation container where the first observation is a binned observation
    """
    # Write header
    if cls._logExplicit():
        cls._log.header3('Binning events')

    # Bin events
    cntcube = ctools.ctbin(obs)
    cntcube['usepnt']   = False
    cntcube['ebinalg']  = 'LOG'
    cntcube['xref']     = cls['xref'].real()
    cntcube['yref']     = cls['yref'].real()
    cntcube['binsz']    = cls['binsz'].real()
    cntcube['nxpix']    = cls['nxpix'].integer()
    cntcube['nypix']    = cls['nypix'].integer()
    cntcube['enumbins'] = cls['enumbins'].integer()
    cntcube['emin']     = cls['emin'].real()
    cntcube['emax']     = cls['emax'].real()
    cntcube['coordsys'] = cls['coordsys'].string()
    cntcube['proj']     = cls['proj'].string()
    cntcube['nthreads'] = nthreads
    cntcube.run()

    # Store counts cube so that we can use it to build the background
    # cube
    cube = cntcube.cube()

    # Write header
    if cls._logExplicit():
        cls._log.header3('Creating stacked response')

    # Get stacked response
    response = get_stacked_response(obs, cube, edisp=cls['edisp'].boolean())

    # Retrieve a new oberservation container
    new_obs = cntcube.obs().copy()

    # Set stacked response
    if cls['edisp'].boolean():
        new_obs[0].response(response['expcube'], response['psfcube'],
                            response['edispcube'], response['bkgcube'])
    else:
        new_obs[0].response(response['expcube'], response['psfcube'],
                            response['bkgcube'])

    # Get new models
    models = response['models']

    # Set models for new oberservation container
    new_obs.models(models)

    # Return new oberservation container
    return new_obs


# ================================ #
# Get On/Off observation container #
# ================================ #
def get_onoff_obs(cls, obs, nthreads=0):
    """
    Create On/Off observations container from given observations

    Parameters
    ----------
    cls : `~ctools.cscript`
        cscript class
    obs : `~gammalib.GObservations`
        Observation container
    nthreads : str, optional
        Number of parallel processes for On/Off spectra computation (0=all available CPUs)

    Returns
    -------
    onoff_obs : `~gammalib.GObservations`
        Observation container with On/Off observations
    """
    # Write header
    if cls._logExplicit():
        cls._log.header3('Creating On/Off observations')

    # Initialise inmodel, srcname, inexclusion and use_model_bkg
    inmodel       = 'NONE'
    srcname       = ''
    inexclusion   = 'NONE'
    use_model_bkg = True

    # Set inmodel, srcname, inexclusion and use_model_bkg if they are available
    if cls.has_par('inmodel') and obs.models().size() == 0:
        if cls['inmodel'].is_valid():
            inmodel = cls['inmodel'].value()
    if cls.has_par('srcname'):
        srcname = cls['srcname'].value()
    if cls.has_par('inexclusion'):
        if cls['inexclusion'].is_valid():
            inexclusion = cls['inexclusion'].value()
    if cls.has_par('use_model_bkg'):
        use_model_bkg = cls['use_model_bkg'].boolean()

    # Create On/Off observations
    phagen = cscripts.csphagen(obs)
    phagen['inmodel']     = inmodel
    phagen['srcname']     = srcname
    phagen['inexclusion'] = inexclusion
    phagen['emin']        = cls['emin'].real()
    phagen['emax']        = cls['emax'].real()
    phagen['enumbins']    = cls['enumbins'].integer()
    phagen['ebinalg']     = 'LOG'
    phagen['srcshape']    = cls['srcshape'].string()
    phagen['coordsys']    = cls['coordsys'].string()
    if cls['coordsys'].string() == 'CEL':
        phagen['ra']  = cls['xref'].real()
        phagen['dec'] = cls['yref'].real()
    elif cls['coordsys'].string() == 'GAL':
        phagen['glon'] = cls['xref'].real()
        phagen['glat'] = cls['yref'].real()
    if cls['srcshape'].string() == 'CIRCLE':
        phagen['rad'] = cls['rad'].real()
    phagen['bkgmethod'] = cls['bkgmethod'].string()
    if cls['bkgmethod'].string() == 'REFLECTED':
        phagen['bkgregmin'] = cls['bkgregmin'].integer()
    phagen['use_model_bkg'] = use_model_bkg
    phagen['maxoffset']     = cls['maxoffset'].real()
    phagen['stack']         = True
    phagen['etruemin']      = cls['etruemin'].real()
    phagen['etruemax']      = cls['etruemax'].real()
    phagen['etruebins']     = cls['etruebins'].integer()
    phagen['chatter']       = cls['chatter'].integer()
    phagen['clobber']       = cls['clobber'].boolean()
    phagen['debug']         = cls['debug'].boolean()
    phagen['nthreads']      = nthreads

    # Pipe exclusion map
    if hasattr(cls, 'exclusion_map'):
        exclusion_map = cls.exclusion_map()
        if exclusion_map is not None:
            phagen.exclusion_map(exclusion_map)

	# Run csphagen
    phagen.run()

    # Clone resulting observation container
    onoff_obs = phagen.obs().copy()

    # On/Off observations are created with CSTAT as default statistic
    # We will change this to the user choice
    if cls['statistic'].string() != 'DEFAULT':
        for observation in onoff_obs:
            observation.statistic(cls['statistic'].string())

    # Return On/Off oberservation container
    return onoff_obs


# ========================================= #
# Calculate residuals from counts and model #
# ========================================= #
def residuals(cls, counts, model):
    """
    Calculate residuals given counts and models, according to algorithm
    specified by user.

    Can handle GSkyMap or GNdarray objects


    Parameters
    ----------
    cls : `~ctools.cscript`
        cscript class
    counts : `~gammalib.GSkyMap/~gammalib.GNdarray'
        Data counts
    model : `~gammalib.GSkyMap/~gammalib.GNdarray'
        Model counts

    Returns
    -------
    residuals : `~gammalib.GSkyMap/~gammalib.GNdarray'
        Residuals
    """
    # Find type of objects we are manipulating and set size to iterate later

    # If GNdarray
    if counts.classname() == 'GNdarray':
        nelem = counts.size()
    elif counts.classname() == 'GSkyMap':
        nelem = counts.npix()
    else:
        msg = 'cscripts.obsutils.residuals only handles ' \
              + 'gammalib.GNdarray or gammalib.GSkyMap objects.\n'
        raise RuntimeError(msg)

    # Copy counts to initialise residuals
    residuals = counts.copy()

    # Get residual map algorithm type
    algorithm = cls['algorithm'].string()

    # Subtract
    if algorithm == 'SUB':
        residuals -= model

    # Subtract and divide by model
    elif algorithm == 'SUBDIV':
        residuals -= model
        residuals /= model

    # Subtract and divide by sqrt of model
    elif algorithm == 'SUBDIVSQRT':
        residuals -= model
        residuals /= model.sqrt()

    # Calculate significance from Li&Ma
    elif algorithm == 'SIGNIFICANCE':

        # Compute sign
        sign = (residuals - model).sign()

        # Loop over every bin
        for i in range(nelem):

            # If the model value > 0.0 do the computation as normal ...
            model_val = model[i]
            if model_val > 0.0:

                # If the data value is also > 0 then compute the
                # significance^2 and save it ...
                data_val = residuals[i]
                if data_val > 0.0:
                    log_val      = math.log(data_val / model_val)
                    residuals[i] = (data_val * log_val) + model_val - data_val

                # ... otherwise compute the reduced value of the above
                # expression. This is necessary to avoid computing log(0).
                else:
                    residuals[i] = model_val

            # ... otherwise hard-code the significance to 0
            else:
                residuals[i] = 0.0

        # Compute significance map
        residuals *= 2.0
        residuals  = residuals.sqrt()
        residuals *= sign

    # Raise exception if algorithm is unknown
    else:
        raise TypeError('Algorithm "' + algorithm + '" not known')

    # Return
    return residuals


# ================= #
# Create counts map #
# ================= #
def create_counts_cube(cls, obs):
    """
    Create counts cube from observations

    Parameters
    ----------
    cls : `~ctools.cscript`
        cscript class
    obs : `~gammalib.GObservations`
        Observation container

    Returns
    -------
    cntcube : `~gammalib.GCTAEventCube`
        Counts cube
    """
    # Initialise ctbin
    ctbin = ctools.ctbin(obs)
    
    # Set parameters
    ctbin['xref']     = cls['xref'].real()
    ctbin['yref']     = cls['yref'].real()
    ctbin['proj']     = cls['proj'].string()
    ctbin['coordsys'] = cls['coordsys'].string()
    ctbin['ebinalg']  = cls['ebinalg'].string()
    ctbin['nxpix']    = cls['nxpix'].integer()
    ctbin['nypix']    = cls['nypix'].integer()
    ctbin['binsz']    = cls['binsz'].real()
    if cls['ebinalg'].string() == 'FILE':
        ctbin['ebinfile'] = cls['ebinfile'].filename().file()
    else:
        ctbin['enumbins'] = cls['enumbins'].integer()
        ctbin['emin']     = cls['emin'].real()
        ctbin['emax']     = cls['emax'].real()
    ctbin['chatter']  = cls['chatter'].integer()
    ctbin['clobber']  = cls['clobber'].boolean()
    ctbin['debug']    = cls['debug'].boolean()

    # Run ctbin
    ctbin.run()

    # Retrieve counts cube
    cntcube = ctbin.cube().copy()

    # Return counts cube
    return cntcube
